from django.utils.translation import ugettext_lazy as _

from rest_framework.reverse import reverse

from mayan.apps.rest_api import serializers

from .models import Source


class SourceSerializer(serializers.HyperlinkedModelSerializer):
    actions = serializers.SerializerMethodField(
        label=_('Actions')
    )

    class Meta:
        extra_kwargs = {
            'url': {
                'label': _('URL'), 'lookup_field': 'pk',
                'lookup_url_kwarg': 'source_id',
                'view_name': 'rest_api:source-detail'
            }
        }
        fields = (
            'actions', 'backend_data', 'backend_path', 'enabled', 'id',
            'label', 'url'
        )
        model = Source
        read_only_fields = ('actions', 'id', 'url')

    def get_actions(self, instance):
        result = []

        for action in instance.get_actions():
            result.append(
                {
                    'name': action.name,
                    'accept_files': action.accept_files,
                    'arguments': action.arguments,
                    'url': reverse(
                        viewname='rest_api:source-action', kwargs={
                            'source_id': instance.pk,
                            'action_name': action.name
                        }, request=self.context['request'],
                        format=self.context['format']
                    )
                }
            )

        return result


class SourceBackendActionSerializer(serializers.Serializer):
    arguments = serializers.JSONField(
        help_text=_(
            'Optional arguments for the action. Must be JSON formatted.'
        ), label=_('Arguments'), required=False
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        action = self.context.get('action')

        if action and action.accept_files:
            self.fields['file'] = serializers.FileField(
                help_text=_('Binary content for the new file.'),
                use_url=False, write_only=True
            )



from django.db import migrations


def code_document_file_size_copy(apps, schema_editor):
    DocumentFile = apps.get_model(
        app_label='documents', model_name='DocumentFile'
    )

    for document_file in DocumentFile.objects.all():
        name = document_file.file.name
        document_file.file.close()

        if document_file.file.storage.exists(name=name):
            document_file.size = document_file.file.storage.size(name=name)
            document_file.save(update_fields=('size',))


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0079_documentfile_size')
    ]

    operations = [
        migrations.RunPython(
            code=code_document_file_size_copy,
            reverse_code=migrations.RunPython.noop
        )
    ]



from django.db.utils import IntegrityError

from mayan.apps.documents.models.trashed_document_models import TrashedDocument
from mayan.apps.documents.tests.base import (
    GenericDocumentTestCase, GenericTransactionDocumentTestCase
)
from mayan.apps.documents.tests.literals import (
    TEST_DOCUMENT_DESCRIPTION, TEST_DOCUMENT_DESCRIPTION_EDITED,
    TEST_DOCUMENT_LABEL_EDITED
)
from mayan.apps.metadata.models import MetadataType, DocumentTypeMetadataType

from ..models import (
    IndexInstance, IndexInstanceNode, IndexTemplate, IndexTemplateNode
)

from .literals import (
    TEST_INDEX_TEMPLATE_DOCUMENT_DESCRIPTION_EXPRESSION,
    TEST_INDEX_TEMPLATE_DOCUMENT_LABEL_EXPRESSION,
    TEST_INDEX_TEMPLATE_DOCUMENT_TYPE_EXPRESSION
)
from .mixins import IndexTemplateTestMixin


class IndexTemplateTestCase(IndexTemplateTestMixin, GenericDocumentTestCase):
    auto_upload_test_document = False

    def test_method_get_absolute_url(self):
        self._create_test_index_template()

        self._clear_events()

        self.assertTrue(self._test_index_template.get_absolute_url())

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class IndexInstanceBasicTestCase(
    IndexTemplateTestMixin, GenericDocumentTestCase
):
    _test_index_template_node_expression = TEST_INDEX_TEMPLATE_DOCUMENT_LABEL_EXPRESSION
    auto_upload_test_document = False

    def test_index_instance_on_document_creation(self):
        self._create_test_document_stub()

        self.assertTrue(
            IndexInstanceNode.objects.filter(
                documents=self._test_document,
            ).exists()
        )

        test_document_id = self._test_document.pk
        self._test_document.delete()

        self.assertTrue(
            IndexInstanceNode.objects.filter(
                documents=test_document_id
            ).exists()
        )

        TrashedDocument.objects.get(
            pk=self._test_document.pk
        ).restore(user=self._test_case_user)
        self._test_document.refresh_from_db()

        self.assertTrue(
            IndexInstanceNode.objects.filter(
                documents=test_document_id
            ).exists()
        )

        self._test_document.delete()
        self._test_document.delete()

        self.assertFalse(
            IndexInstanceNode.objects.filter(
                documents=test_document_id
            ).exists()
        )

    def test_index_instance_node_id_persistence(self):
        self._create_test_document_stub()

        index_instance_node_id = IndexInstanceNode.objects.get(
            documents=self._test_document
        ).pk

        IndexInstance.objects.document_add(document=self._test_document)

        self.assertEqual(
            IndexInstanceNode.objects.get(documents=self._test_document).pk,
            index_instance_node_id
        )

    def test_index_instance_node_cleanup(self):
        self._create_test_document_stub()

        document_index_instance_node_count = IndexInstanceNode.objects.count()

        self._test_document.label = TEST_DOCUMENT_LABEL_EDITED
        self._test_document.save()

        self.assertEqual(
            IndexInstanceNode.objects.count(),
            document_index_instance_node_count
        )


class IndexInstanceTestCase(IndexTemplateTestMixin, GenericDocumentTestCase):
    auto_upload_test_document = False
    auto_create_test_index_template_node = False

    def setUp(self):
        super().setUp()
        self._create_test_document_stub()

    def test_date_based_index(self):
        level_year = self._test_index_template.index_template_nodes.create(
            parent=self._test_index_template.index_template_root_node,
            expression='{{ document.datetime_created|date:"Y" }}',
            link_documents=False
        )

        self._test_index_template.index_template_nodes.create(
            parent=level_year,
            expression='{{ document.datetime_created|date:"m" }}',
            link_documents=True
        )
        # Index the document created by default.
        IndexTemplate.objects.rebuild()

        self._test_document.delete()

        # Uploading a new should not trigger an error.
        self._upload_test_document()

        self.assertEqual(
            list(IndexInstanceNode.objects.values_list('value', flat=True)),
            [
                '', str(self._test_document.datetime_created.year),
                '{:02}'.format(self._test_document.datetime_created.month)
            ]
        )

        self.assertTrue(
            self._test_document in IndexInstanceNode.objects.last().documents.all()
        )

    def test_dual_level_dual_document_index(self):
        """
        Test creation of an index instance with two first levels with different
        values and two second levels with the same value but as separate
        children of each of the first levels. GitLab issue #391
        """
        self._create_test_document_stub()

        # Create simple index template
        level_1 = self._test_index_template.index_template_nodes.create(
            expression='{{ document.uuid }}', link_documents=False,
            parent=self._test_index_template_root_node
        )

        self._test_index_template.index_template_nodes.create(
            expression='{{ document.label }}', link_documents=True,
            parent=level_1
        )

        IndexTemplate.objects.rebuild()

        # Typecast to sets to make sorting irrelevant in the comparison.
        self.assertEqual(
            set(IndexInstanceNode.objects.values_list('value', flat=True)),
            {
                '', str(self._test_documents[1].uuid),
                self._test_documents[1].label,
                str(self._test_documents[0].uuid),
                self._test_documents[0].label
            }
        )

    def test_document_description_index(self):
        self._create_test_index_template_node(
            expression=TEST_INDEX_TEMPLATE_DOCUMENT_DESCRIPTION_EXPRESSION
        )

        self._test_document.description = TEST_DOCUMENT_DESCRIPTION
        self._test_document.save()

        self._test_index_template.rebuild()

        self.assertEqual(
            IndexInstanceNode.objects.last().value,
            self._test_document.description
        )
        self._test_document.description = TEST_DOCUMENT_DESCRIPTION_EDITED
        self._test_document.save()

        self.assertEqual(
            IndexInstanceNode.objects.last().value,
            self._test_document.description
        )

    def test_document_label_index(self):
        self._create_test_index_template_node(
            expression=TEST_INDEX_TEMPLATE_DOCUMENT_LABEL_EXPRESSION
        )

        self._test_index_template.rebuild()

        self.assertEqual(
            IndexInstanceNode.objects.last().value, self._test_document.label
        )
        self._test_document.label = TEST_DOCUMENT_LABEL_EDITED
        self._test_document.save()

        self.assertEqual(
            IndexInstanceNode.objects.last().value, self._test_document.label
        )

    def test_document_type_index(self):
        self._create_test_index_template_node(
            expression=TEST_INDEX_TEMPLATE_DOCUMENT_TYPE_EXPRESSION
        )

        self._test_index_template.rebuild()

        self.assertEqual(
            IndexInstanceNode.objects.last().value,
            self._test_document_types[0].label
        )

        self._create_test_document_type()
        self._test_index_template.document_types.add(self._test_document_type)

        self._test_document._document_type_change(
            document_type=self._test_document_type
        )

        self.assertEqual(
            IndexInstanceNode.objects.last().value,
            self._test_document_types[1].label
        )

    def test_method_get_absolute_url(self):
        test_index_instance = IndexInstance.objects.first()
        self.assertTrue(test_index_instance.get_absolute_url())

    def test_multi_level_template_with_no_result_parent(self):
        """
        On a two level template if the first level doesn't return a result
        the indexing should stop. GitLab issue #391.
        """
        level_1 = self._test_index_template.index_template_nodes.create(
            expression='', link_documents=True,
            parent=self._test_index_template_root_node,
        )

        self._test_index_template.index_template_nodes.create(
            expression='{{ document.label }}', link_documents=True,
            parent=level_1
        )

        IndexTemplate.objects.rebuild()

    def test_rebuild_all_indexes(self):
        # Add metadata type and connect to document type.
        metadata_type = MetadataType.objects.create(
            name='test', label='test'
        )
        DocumentTypeMetadataType.objects.create(
            document_type=self._test_document_type,
            metadata_type=metadata_type
        )

        # Add document metadata value.
        self._test_document.metadata.create(
            metadata_type=metadata_type, value='0001'
        )

        self._create_test_index_template_node(
            expression='{{ document.metadata_value_of.test }}'
        )

        self.assertEqual(
            list(
                IndexTemplateNode.objects.values_list('expression', flat=True)
            ), ['', '{{ document.metadata_value_of.test }}']
        )

        # There should be only a root index instances nodes.
        self.assertEqual(IndexInstanceNode.objects.count(), 1)
        self.assertEqual(IndexInstanceNode.objects.first().parent, None)

        # Rebuild all indexes.
        IndexTemplate.objects.rebuild()

        # Check that document is in instance node.
        instance_node = IndexInstanceNode.objects.get(value='0001')
        self.assertQuerysetEqual(
            instance_node.documents.all(), [repr(self._test_document)]
        )


class IndexIntegrityTestCase(
    IndexTemplateTestMixin, GenericTransactionDocumentTestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_document_stub()

    def test_unique_value_per_level(self):
        self._create_test_index_template_node(
            expression=TEST_INDEX_TEMPLATE_DOCUMENT_LABEL_EXPRESSION
        )

        self._test_index_template.rebuild()

        index_instance_node = IndexInstanceNode.objects.last()

        with self.assertRaises(expected_exception=IntegrityError):
            IndexInstanceNode.objects.create(
                index_template_node=index_instance_node.index_template_node,
                parent=index_instance_node.parent,
                value=index_instance_node.value
            )

        # Reset the failed database write to allow the database manager
        # to flush the database during the test tear down.
        IndexInstanceNode.objects.create(
            parent=index_instance_node.parent,
            index_template_node=index_instance_node.index_template_node,
            value='{}_{}'.format(
                index_instance_node.value, index_instance_node.pk
            )
        )



from django.db import models
from django.urls import reverse
from django.utils.translation import ugettext_lazy as _

from ..managers import DocumentFilePageManager, ValidDocumentFilePageManager

from .document_file_models import DocumentFile
from .document_file_page_model_mixins import DocumentFilePageBusinessLogicMixin
from .model_mixins import PagedModelMixin

__all__ = ('DocumentFilePage', 'DocumentFilePageSearchResult')


class DocumentFilePage(
    DocumentFilePageBusinessLogicMixin, PagedModelMixin, models.Model
):
    """
    Model that describes a document file page
    """
    _paged_model_parent_field = 'document_file'

    document_file = models.ForeignKey(
        on_delete=models.CASCADE, related_name='file_pages', to=DocumentFile,
        verbose_name=_('Document file')
    )
    page_number = models.PositiveIntegerField(
        db_index=True, default=1, editable=False,
        verbose_name=_('Page number')
    )

    objects = DocumentFilePageManager()
    valid = ValidDocumentFilePageManager()

    class Meta:
        ordering = ('page_number',)
        verbose_name = _('Document file page')
        verbose_name_plural = _('Document file pages')

    def __str__(self):
        return self.get_label()

    def delete(self, *args, **kwargs):
        self.cache_partition.delete()
        super().delete(*args, **kwargs)

    def get_absolute_url(self):
        return reverse(
            viewname='documents:document_file_page_view', kwargs={
                'document_file_page_id': self.pk
            }
        )

    def natural_key(self):
        return (
            self.page_number, self.document_file.natural_key()
        )
    natural_key.dependencies = ['documents.DocumentFile']


class DocumentFilePageSearchResult(DocumentFilePage):
    class Meta:
        ordering = ('document_file__document', 'page_number')
        proxy = True
        verbose_name = _('Document file page')
        verbose_name_plural = _('Document file pages')



from django.utils.translation import ugettext_lazy as _

DOCUMENT_FILE_DOWNLOAD_MESSAGE_BODY = _(
    'The document files have compressed '
    'and are available for download using the '
    'link: %(download_url)s or from '
    'the downloads area (%(download_list_url)s).'
)
DOCUMENT_FILE_DOWNLOAD_MESSAGE_SUBJECT = _(
    'Document files ready for download.'
)



from django import forms as django_forms
from django.utils.module_loading import import_string


class DynamicFormMixin:
    def __init__(self, *args, **kwargs):
        self.schema = kwargs.pop('schema')
        super().__init__(*args, **kwargs)

        widgets = self.schema.get(
            'widgets', {}
        )
        field_order = self.schema.get(
            'field_order', self.schema['fields'].keys()
        )

        for field_name in field_order:
            field_data = self.schema['fields'][field_name]
            field_class = import_string(
                dotted_path=field_data['class']
            )
            kwargs = {
                'label': field_data['label'],
                'required': field_data.get('required', True),
                'initial': field_data.get('default', None),
                'help_text': field_data.get('help_text')
            }
            if widgets and field_name in widgets:
                widget = widgets[field_name]
                kwargs['widget'] = import_string(
                    dotted_path=widget['class']
                )(
                    **widget.get(
                        'kwargs', {}
                    )
                )

            kwargs.update(
                field_data.get(
                    'kwargs', {}
                )
            )
            self.fields[field_name] = field_class(**kwargs)

    @property
    def media(self):
        """
        Append the media of the dynamic fields to the normal fields' media.
        """
        media = super().media
        media += django_forms.Media(
            **self.schema.get(
                'media', {}
            )
        )
        return media


class FormFieldsetMixin:
    fieldsets = None

    def get_fieldsets(self):
        if self.fieldsets:
            return self.fieldsets
        else:
            return (
                (
                    None, {
                        'fields': tuple(self.fields)
                    }
                ),
            )



from django.urls import reverse

from mayan.apps.testing.tests.base import BaseTestCase

from ..models import DownloadFile, SharedUploadedFile
from ..settings import (
    setting_download_file_expiration_interval,
    setting_shared_uploaded_file_expiration_interval
)

from .mixins import DownloadFileTestMixin, SharedUploadedFileTestMixin


class DownloadFileModelTestCase(DownloadFileTestMixin, BaseTestCase):
    def test_download_file_expiration(self):
        setting_download_file_expiration_interval.set(value=60)
        self._create_test_download_file()

        self.assertEqual(DownloadFile.objects.stale().count(), 0)

        setting_download_file_expiration_interval.set(value=0)

        self.assertEqual(DownloadFile.objects.stale().count(), 1)

    def test_method_get_absolute_url(self):
        self._create_test_download_file()

        self.assertEqual(
            self.test_download_file.get_absolute_url(),
            reverse(viewname='storage:download_file_list')
        )


class SharedUploadedFileManagerTestCase(
    SharedUploadedFileTestMixin, BaseTestCase
):
    def test_shared_uploaded_expiration(self):
        setting_shared_uploaded_file_expiration_interval.set(value=60)
        self._create_test_shared_uploaded_file()

        self.assertEqual(SharedUploadedFile.objects.stale().count(), 0)

        setting_shared_uploaded_file_expiration_interval.set(value=0)

        self.assertEqual(SharedUploadedFile.objects.stale().count(), 1)



import json
import os

from django import forms as django_forms
from django.apps import apps
from django.conf import settings
from django.contrib.admin.utils import (
    get_fields_from_path, help_text_for_field, label_for_field
)
from django.core.exceptions import FieldDoesNotExist, ImproperlyConfigured
from django.db import models
from django.db.models import Model
from django.db.models.query import QuerySet
from django.forms import Form as DjangoForm, ModelForm as DjangoModelForm
from django.forms.models import ModelFormMetaclass
from django.utils.translation import ugettext_lazy as _

from mayan.apps.common.utils import resolve_attribute

from .form_mixins import DynamicFormMixin, FormFieldsetMixin
from .widgets import DisableableSelectWidget, PlainWidget, TextAreaDiv


class Form(FormFieldsetMixin, DjangoForm):
    """Mayan's default form class."""


class ModelForm(FormFieldsetMixin, DjangoModelForm):
    """Mayan's default model form class."""


class ChoiceForm(Form):
    """
    Form to be used in side by side templates used to add or remove
    items from a many to many field.
    """
    search = django_forms.CharField(
        label=_('Search'), required=False,
        widget=django_forms.widgets.TextInput(
            attrs={
                'autocomplete': 'off',
                'class': 'views-select-search',
                'placeholder': 'Filter list'
            }
        )
    )
    selection = django_forms.MultipleChoiceField(
        required=False, widget=DisableableSelectWidget()
    )

    def __init__(self, *args, **kwargs):
        choices = kwargs.pop('choices', [])
        label = kwargs.pop(
            'label', _('Selection')
        )
        help_text = kwargs.pop('help_text', None)
        disabled_choices = kwargs.pop(
            'disabled_choices', ()
        )
        super().__init__(*args, **kwargs)
        self.fields['selection'].choices = choices
        self.fields['selection'].label = label
        self.fields['selection'].help_text = help_text
        self.fields['selection'].widget.disabled_choices = disabled_choices
        self.fields['selection'].widget.attrs.update(
            {
                'class': 'full-height input-hotkey-double-click',
                'data-height-difference': '495'
            }
        )


class FormOptions:
    def __init__(self, form, kwargs, options=None):
        """
        Option definitions will be iterated. The option value will be
        determined in the following order: as passed via keyword
        arguments during form intialization, as form get_... method or
        finally as static Meta options. This is to allow a form with
        Meta options or method to be overridden at initialization
        and increase the usability of a single class.
        """
        for name, default_value in self.option_definitions.items():
            try:
                # Check for a runtime value via kwargs
                value = kwargs.pop(name)
            except KeyError:
                try:
                    # Check if there is a get_... method
                    value = getattr(
                        self, 'get_{}'.format(name)
                    )()
                except AttributeError:
                    try:
                        # Check the meta class options
                        value = getattr(options, name)
                    except AttributeError:
                        value = default_value

            setattr(self, name, value)


class DetailFormOption(FormOptions):
    # Dictionary list of option names and default values.
    option_definitions = {
        'extra_fields': []
    }


class DetailForm(ModelForm):
    def __init__(self, *args, **kwargs):
        self.opts = DetailFormOption(
            form=self, kwargs=kwargs, options=getattr(self, 'Meta', None)
        )
        super().__init__(*args, **kwargs)

        for field_index, extra_field in enumerate(iterable=self.opts.extra_fields):
            obj = extra_field.get('object', self.instance)
            field = extra_field.get('field', None)
            func = extra_field.get('func', None)
            label = extra_field.get('label', None)
            help_text = extra_field.get('help_text', None)

            if field:
                if not label:
                    # If label is not specified try to get it from the object
                    # itself.
                    try:
                        fields = get_fields_from_path(model=obj, path=field)
                    except FieldDoesNotExist:
                        # Might be property of a method.
                        label = getattr(
                            getattr(obj, field), 'short_description',
                            field
                        )
                    else:
                        label = label_for_field(
                            model=obj, name=fields[-1].name
                        )

                if not help_text:
                    # If help_text is not specified try to get it from the
                    # object itself.
                    try:
                        fields = get_fields_from_path(model=obj, path=field)
                    except FieldDoesNotExist:
                        # Might be property of a method.
                        help_text = getattr(
                            getattr(obj, field), 'help_text', None
                        )
                    else:
                        help_text = help_text_for_field(
                            model=obj, name=fields[-1].name
                        )

                value = resolve_attribute(attribute=field, obj=obj)

            if func:
                value = func(obj)

            field = field or 'anonymous_field_{}'.format(field_index)

            if isinstance(value, models.query.QuerySet):
                self.fields[field] = django_forms.ModelMultipleChoiceField(
                    queryset=value, label=label
                )
            else:
                self.fields[field] = django_forms.CharField(
                    initial=value, label=label, help_text=help_text,
                    widget=extra_field.get('widget', PlainWidget)
                )

        for field_name, field in self.fields.items():
            self.fields[field_name].widget.attrs.update(
                {'readonly': 'readonly'}
            )


class DynamicForm(DynamicFormMixin, Form):
    """Normal dynamic form."""


class DynamicModelForm(DynamicFormMixin, ModelForm):
    """Dynamic model form."""


class DynamicFormMetaclass(ModelFormMetaclass):
    def __new__(mcs, name, bases, attrs):
        new_class = super(DynamicFormMetaclass, mcs).__new__(
            mcs=mcs, name=name, bases=bases, attrs=attrs
        )

        if new_class._meta.fields:
            new_class._meta.fields += ('backend_data',)
            widgets = getattr(
                new_class._meta, 'widgets', {}
            ) or {}
            widgets['backend_data'] = django_forms.widgets.HiddenInput
            new_class._meta.widgets = widgets

        return new_class


class BackendDynamicForm(DynamicModelForm, metaclass=DynamicFormMetaclass):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        backend_data = self.instance.get_backend_data()

        if backend_data:
            for field_name in self.instance.get_backend().get_fields():
                self.fields[field_name].initial = backend_data.get(
                    field_name, None
                )

    def clean(self):
        data = super().clean()

        # Consolidate the dynamic fields into a single JSON field called
        # 'backend_data'.
        backend_data = {}

        for field_name, field_data in self.schema['fields'].items():
            backend_data[field_name] = data.pop(
                field_name, field_data.get('default', None)
            )
            if isinstance(backend_data[field_name], QuerySet):
                # Flatten the queryset to a list of ids.
                backend_data[field_name] = list(
                    backend_data[field_name].values_list('id', flat=True)
                )
            elif isinstance(backend_data[field_name], Model):
                # Store only the ID of a model instance.
                backend_data[field_name] = backend_data[field_name].pk

        data['backend_data'] = json.dumps(obj=backend_data)

        return data


class FileDisplayForm(Form):
    DIRECTORY = None
    FILENAME = None

    text = django_forms.CharField(
        label='',
        widget=TextAreaDiv(
            attrs={
                'class': 'full-height scrollable',
                'data-height-difference': 270
            }
        )
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.DIRECTORY or self.FILENAME:
            file_path = os.path.join(
                settings.BASE_DIR, os.sep.join(self.DIRECTORY), self.FILENAME
            )
            with open(file=file_path) as file_object:
                self.fields['text'].initial = file_object.read()


class FilteredSelectionFormOptions(FormOptions):
    # Dictionary list of option names and default values.
    option_definitions = {
        'allow_multiple': False,
        'field_name': None,
        'help_text': None,
        'label': None,
        'model': None,
        'permission': None,
        'queryset': None,
        'required': True,
        'user': None,
        'widget_attributes': {'size': '10'},
        'widget_class': None
    }


class FilteredSelectionForm(Form):
    """
    Form to select the from a list of choice filtered by access. Can be
    configure to allow single or multiple selection.
    """
    def __init__(self, *args, **kwargs):
        opts = FilteredSelectionFormOptions(
            form=self, kwargs=kwargs, options=getattr(self, 'Meta', None)
        )

        if opts.queryset is None:
            if not opts.model:
                raise ImproperlyConfigured(
                    '{} requires a queryset or a model to be specified as '
                    'a meta option or passed during initialization.'.format(
                        self.__class__.__name__
                    )
                )

            queryset = opts.model.objects.all()
        else:
            queryset = opts.queryset

        if opts.allow_multiple:
            extra_kwargs = {}
            field_class = django_forms.ModelMultipleChoiceField
            widget_class = django_forms.widgets.SelectMultiple
        else:
            extra_kwargs = {'empty_label': None}
            field_class = django_forms.ModelChoiceField
            widget_class = django_forms.widgets.Select

        if opts.widget_class:
            widget_class = opts.widget_class

        if opts.permission:
            AccessControlList = apps.get_model(
                app_label='acls', model_name='AccessControlList'
            )
            queryset = AccessControlList.objects.restrict_queryset(
                permission=opts.permission, queryset=queryset,
                user=opts.user
            )

        super().__init__(*args, **kwargs)

        self.fields[opts.field_name] = field_class(
            help_text=opts.help_text, label=opts.label,
            queryset=queryset, required=opts.required,
            widget=widget_class(attrs=opts.widget_attributes),
            **extra_kwargs
        )


class RelationshipForm(Form):
    def __init__(self, *args, **kwargs):
        self._event_actor = kwargs.pop('_event_actor')
        super().__init__(*args, **kwargs)

        self.fields['label'] = django_forms.CharField(
            label=_('Label'), required=False,
            widget=django_forms.TextInput(
                attrs={'readonly': 'readonly'}
            )
        )
        self.fields['relationship_type'] = django_forms.ChoiceField(
            choices=self.RELATIONSHIP_CHOICES, label=_('Relationship'),
            widget=django_forms.RadioSelect()
        )

        self.sub_object = self.initial.get('sub_object')
        if self.sub_object:
            self.fields['label'].initial = str(self.sub_object)

            self.initial_relationship_type = self.get_relationship_type()

            self.fields['relationship_type'].initial = self.initial_relationship_type

    def get_new_relationship_instance(self):
        related_manager = getattr(
            self.initial.get('object'),
            self.initial['relationship_related_field']
        )
        main_field_name = related_manager.field.name

        return related_manager.model(
            **{
                main_field_name: self.initial.get('object'),
                self.initial['relationship_related_query_field']: self.initial.get('sub_object')
            }
        )

    def get_relationship_queryset(self):
        return getattr(
            self.initial.get('object'),
            self.initial['relationship_related_field']
        ).filter(
            **{
                self.initial['relationship_related_query_field']: self.initial.get('sub_object')
            }
        )

    def get_relationship_instance(self):
        relationship_queryset = self.get_relationship_queryset()
        if relationship_queryset.exists():
            return relationship_queryset.get()
        else:
            return self.get_new_relationship_instance()

    def save(self):
        if self.sub_object:
            if self.cleaned_data['relationship_type'] != self.initial_relationship_type:
                save_method = getattr(
                    self, 'save_relationship_{}'.format(
                        self.cleaned_data['relationship_type']
                    )
                )
                save_method()



from mayan.apps.documents.tests.literals import TEST_FILE_HYBRID_PDF_CONTENT

from ..events import (
    event_parsing_document_file_content_deleted,
    event_parsing_document_file_finished
)
from ..models import DocumentFilePageContent


class DocumentFileContentToolsViewTestMixin:
    def _request_document_type_parsing_view(self):
        return self.post(
            viewname='document_parsing:document_type_submit', data={
                'document_type': self._test_document_type.pk
            }
        )


class DocumentFileContentTestMixin:
    auto_create_test_document_file_parsed_content = False

    def setUp(self):
        super().setUp()
        if self.auto_create_test_document_file_parsed_content:
            self._create_test_document_file_parsed_content()

    def _create_test_document_file_parsed_content(self):
        DocumentFilePageContent.objects.create(
            document_file_page=self._test_document_file_page,
            content=TEST_FILE_HYBRID_PDF_CONTENT
        )
        event_parsing_document_file_finished.commit(
            action_object=self._test_document_file_page.document_file.document,
            target=self._test_document_file_page.document_file
        )

    def _do_test_document_file_parsed_content_delete(self):
        DocumentFilePageContent.objects.delete_content_for(
            document_file=self._test_document_file
        )
        event_parsing_document_file_content_deleted.commit(
            action_object=self._test_document_file.document,
            target=self._test_document_file
        )


class DocumentFileContentViewTestMixin:
    def _request_test_document_file_content_single_delete_view(self):
        return self.post(
            viewname='document_parsing:document_file_content_single_delete',
            kwargs={
                'document_file_id': self._test_document_file.pk
            }
        )

    def _request_test_document_file_content_multiple_delete_view(self):
        return self.post(
            viewname='document_parsing:document_file_content_multiple_delete',
            data={
                'id_list': self._test_document_file.pk
            }
        )

    def _request_test_document_file_content_download_view(self):
        return self.get(
            viewname='document_parsing:document_file_content_download',
            kwargs={
                'document_file_id': self._test_document_file.pk
            }
        )

    def _request_test_document_file_content_view(self):
        return self.get(
            'document_parsing:document_file_content_view', kwargs={
                'document_file_id': self._test_document_file.pk
            }
        )

    def _request_test_document_file_page_content_view(self):
        return self.get(
            viewname='document_parsing:document_file_page_content_view',
            kwargs={
                'document_file_page_id': self._test_document_file.pages.first().pk,
            }
        )

    def _request_test_document_file_parsing_submit_view(self):
        return self.post(
            viewname='document_parsing:document_file_parsing_single_submit', kwargs={
                'document_file_id': self._test_document_file.pk
            }
        )

    def _request_test_document_parsing_submit_view(self):
        return self.post(
            viewname='document_parsing:document_parsing_single_submit', kwargs={
                'document_id': self._test_document.pk
            }
        )


class DocumentFilePageContentAPITestMixin:
    def _request_document_file_page_content_api_view(self):
        return self.get(
            viewname='rest_api:document-file-page-content-view', kwargs={
                'document_id': self._test_document.pk,
                'document_file_id': self._test_document_file.pk,
                'document_file_page_id': self._test_document_file.pages.first().pk
            }
        )


class DocumentTypeContentViewTestMixin:
    def _request_test_document_type_parsing_settings_view(self):
        return self.get(
            viewname='document_parsing:document_type_parsing_settings',
            kwargs={'document_type_id': self._test_document_type.pk}
        )


class DocumentTypeParsingSettingsAPIViewTestMixin():
    def _request_document_type_parsing_settings_details_api_view(self):
        return self.get(
            viewname='rest_api:document-type-parsing-settings-view',
            kwargs={'document_type_id': self._test_document_type.pk}
        )

    def _request_document_type_parsing_settings_patch_api_view(self):
        return self.patch(
            viewname='rest_api:document-type-parsing-settings-view',
            kwargs={'document_type_id': self._test_document_type.pk},
            data={'auto_parsing': True}
        )

    def _request_document_type_parsing_settings_put_api_view(self):
        return self.put(
            viewname='rest_api:document-type-parsing-settings-view',
            kwargs={'document_type_id': self._test_document_type.pk},
            data={'auto_parsing': True}
        )



from django.utils.translation import ugettext_lazy as _

from mayan.apps.task_manager.classes import CeleryQueue
from mayan.apps.task_manager.workers import worker_b

queue_search = CeleryQueue(
    label=_('Search'), name='search', worker=worker_b
)
queue_search_slow = CeleryQueue(
    label=_('Search slow'), name='search_slow', worker=worker_b
)

queue_search.add_task_type(
    dotted_path='mayan.apps.dynamic_search.tasks.task_deindex_instance',
    label=_('Remove a model instance from the search engine.'),
    name='task_deindex_instance'
)
queue_search.add_task_type(
    dotted_path='mayan.apps.dynamic_search.tasks.task_index_instance',
    label=_('Index a model instance to the search engine.'),
    name='task_index_instance'
)
queue_search.add_task_type(
    dotted_path='mayan.apps.dynamic_search.tasks.task_index_instances',
    label=_('Index all instances of a search model to the search engine.'),
    name='task_index_instances'
)
queue_search.add_task_type(
    dotted_path='mayan.apps.dynamic_search.tasks.task_index_related_instance_m2m',
    label=_(
        'Index all related instances of a search model after a many to many '
        'event.'
    ),
    name='task_index_related_instance_m2m'
)

queue_search_slow.add_task_type(
    dotted_path='mayan.apps.dynamic_search.tasks.task_reindex_backend',
    label=_('Reset the search backend indices and index all instances again.'),
    name='task_reindex_backend'
)



from unittest import skip

from mayan.apps.testing.tests.base import BaseTestCase

from ..exceptions import DynamicSearchBackendException
from ..search_query_types import QueryTypeExact

from .mixins.base import SearchTestMixin, TestSearchObjectSimpleTestMixin
from .mixins.backend_mixins import BackendSearchTestMixin
from .mixins.backend_search_field_mixins import BackendSearchFieldTestCaseMixin
from .mixins.backend_query_type_mixins import BackendFieldTypeQueryTypeTestCaseMixin


class DjangoSearchBackendSearchFieldTestCase(
    BackendSearchFieldTestCaseMixin, BaseTestCase
):
    _test_search_backend_path = 'mayan.apps.dynamic_search.backends.django.DjangoSearchBackend'


class DjangoSearchBackendFieldTypeQueryTypeTestCase(
    BackendFieldTypeQueryTypeTestCaseMixin, BaseTestCase
):
    _test_search_backend_path = 'mayan.apps.dynamic_search.backends.django.DjangoSearchBackend'

    @skip(reason='Backend does not support the feature.')
    def test_search_field_type_char_search_exact_accent(self):
        """
        This backend does not require indexing which is required
        for this feature to work.
        """

    @skip(reason='Backend does not support the feature.')
    def test_search_field_type_char_search_fuzzy(self):
        """
        This query type is emulated and does not return the same results
        as backends support this natively.
        """

    @skip(reason='Backend does not support the feature.')
    def test_search_field_type_text_search_fuzzy(self):
        """
        This query type is emulated and does not return the same results
        as backends support this natively.
        """


@skip(reason='Skip until a Mock ElasticSearch server class is added.')
class ElasticSearchBackendIndexingTestCase(
    BackendSearchTestMixin, TestSearchObjectSimpleTestMixin, SearchTestMixin,
    BaseTestCase
):
    _test_search_backend_path = 'mayan.apps.dynamic_search.backends.elasticsearch.ElasticSearchBackend'

    def test_search_without_indexes(self):
        self._test_search_backend.tear_down()

        with self.assertRaises(expected_exception=DynamicSearchBackendException):
            self._do_backend_search(
                field_name='char',
                query_type=QueryTypeExact,
                value=self._test_object.char,
                _skip_refresh=True,
            )


@skip(reason='Skip until a Mock ElasticSearch server class is added.')
class ElasticSearchBackendSearchFieldTestCase(
    BackendSearchFieldTestCaseMixin, BaseTestCase
):
    _test_search_backend_path = 'mayan.apps.dynamic_search.backends.elasticsearch.ElasticSearchBackend'


@skip(reason='Skip until a Mock ElasticSearch server class is added.')
class ElasticSearchBackendFieldTypeQueryTypeTestCase(
    BackendFieldTypeQueryTypeTestCaseMixin, BaseTestCase
):
    _test_search_backend_path = 'mayan.apps.dynamic_search.backends.elasticsearch.ElasticSearchBackend'


class WhooshSearchBackendSearchFieldTestCase(
    BackendSearchFieldTestCaseMixin, BaseTestCase
):
    _test_search_backend_path = 'mayan.apps.dynamic_search.backends.whoosh.WhooshSearchBackend'


class WhooshSearchBackendFieldTypeQueryTypeTestCase(
    BackendFieldTypeQueryTypeTestCaseMixin, BaseTestCase
):
    _test_search_backend_path = 'mayan.apps.dynamic_search.backends.whoosh.WhooshSearchBackend'


class WhooshSearchBackendSpecificTestCase(
    BackendSearchTestMixin, TestSearchObjectSimpleTestMixin, SearchTestMixin,
    BaseTestCase
):
    _test_search_backend_path = 'mayan.apps.dynamic_search.backends.whoosh.WhooshSearchBackend'

    def test_whoosh_datetime_search_raw_parsed_date_human_today(self):
        id_list = self._do_backend_search(
            field_name='datetime',
            is_raw_value=True,
            query_type=QueryTypeExact,
            value='today'
        )

        self.assertEqual(len(id_list), 1)
        self.assertTrue(self._test_object.id in id_list)

    def test_whoosh_datetime_search_raw_parsed_date_human_range(self):
        id_list = self._do_backend_search(
            field_name='datetime',
            is_raw_value=True,
            query_type=QueryTypeExact,
            value='[\'last tuesday\' to \'next friday\']'
        )

        self.assertEqual(len(id_list), 1)
        self.assertTrue(self._test_object.id in id_list)

    def test_whoosh_datetime_search_raw_parsed_date_numeric_range(self):
        id_list = self._do_backend_search(
            field_name='datetime',
            is_raw_value=True,
            query_type=QueryTypeExact,
            value='[\'{}\' to \'{}\']'.format(
                self._test_object.datetime.year - 1,
                self._test_object.datetime.year + 1
            )
        )

        self.assertEqual(len(id_list), 1)
        self.assertTrue(self._test_object.id in id_list)

    def test_whoosh_integer_search_raw_parsed_numeric_range(self):
        id_list = self._do_backend_search(
            field_name='integer',
            is_raw_value=True,
            query_type=QueryTypeExact,
            value='[\'{}\' to \'{}\']'.format(
                self._test_object.integer - 1,
                self._test_object.integer + 1
            )
        )

        self.assertEqual(len(id_list), 1)
        self.assertTrue(self._test_object.id in id_list)



import logging

from django.conf import settings
from django.db import models
from django.utils.translation import ugettext_lazy as _

from mayan.apps.databases.model_mixins import ExtraDataModelMixin

from ..managers import (
    FavoriteDocumentManager, ValidFavoriteDocumentProxyManager,
    ValidFavoriteDocumentManager
)

from .document_models import Document

__all__ = ('FavoriteDocument',)
logger = logging.getLogger(name=__name__)


class FavoriteDocument(ExtraDataModelMixin, models.Model):
    """
    Keeps a list of the favorited documents of a given user.
    """
    user = models.ForeignKey(
        db_index=True, editable=False, on_delete=models.CASCADE,
        to=settings.AUTH_USER_MODEL, verbose_name=_('User')
    )
    document = models.ForeignKey(
        editable=False, on_delete=models.CASCADE, related_name='favorites',
        to=Document, verbose_name=_('Document')
    )
    datetime_added = models.DateTimeField(
        auto_now=True, db_index=True, verbose_name=_('Date and time added')
    )

    objects = FavoriteDocumentManager()
    valid = ValidFavoriteDocumentManager()

    class Meta:
        ordering = ('datetime_added',)
        verbose_name = _('Favorite document')
        verbose_name_plural = _('Favorite documents')

    def __str__(self):
        return str(self.document)

    def natural_key(self):
        return (
            self.document.natural_key(), self.user.natural_key()
        )
    natural_key.dependencies = [
        'documents.Document', settings.AUTH_USER_MODEL
    ]


class FavoriteDocumentProxy(Document):
    objects = models.Manager()
    valid = ValidFavoriteDocumentProxyManager()

    class Meta:
        proxy = True



from django.contrib import admin

from .models import UserLocaleProfile


@admin.register(UserLocaleProfile)
class UserLocaleProfileAdmin(admin.ModelAdmin):
    list_display = ('user', 'timezone', 'language',)
    list_filter = ('timezone', 'language')
    search_fields = (
        'user__username', 'user__first_name', 'user__last_name',
        'user__email', 'timezone'
    )



from django.utils.translation import ugettext_lazy as _

from mayan.apps.documents.queues import queue_documents

queue_documents.add_task_type(
    dotted_path='mayan.apps.document_exports.tasks.task_document_version_export',
    label=_('Export a document version to a PDF')
)



from datetime import timedelta

from django.db import models
from django.utils.timezone import now

from .settings import (
    setting_download_file_expiration_interval,
    setting_shared_uploaded_file_expiration_interval
)


class DownloadFileManager(models.Manager):
    def stale(self):
        return self.filter(
            datetime__lt=now() - timedelta(
                seconds=setting_download_file_expiration_interval.value
            )
        )


class SharedUploadedFileManager(models.Manager):
    def stale(self):
        return self.filter(
            datetime__lt=now() - timedelta(
                seconds=setting_shared_uploaded_file_expiration_interval.value
            )
        )



import json

from mayan.apps.documents.tests.literals import TEST_FILE_SMALL_PATH

from ...source_backends.literals import SOURCE_UNCOMPRESS_CHOICE_NEVER
from ...source_backends.web_form_backends import SourceBackendWebForm

from .base_mixins import SourceTestMixin


class WebFormSourceBackendAPITestMixin:
    def _request_test_web_form_file_upload_action_api_view(self, test_file_path=TEST_FILE_SMALL_PATH):
        with open(file=test_file_path, mode='rb') as file_object:
            return self.post(
                viewname='rest_api:source-action', kwargs={
                    'action_name': 'upload', 'source_id': self._test_source.pk
                }, data={
                    'arguments': json.dumps(
                        obj={
                            'document_type_id': self._test_document_type.pk,
                            'expand': True
                        }
                    ), 'file': file_object
                }
            )


class WebFormSourceBackendTestMixin(SourceTestMixin):
    _create_source_method = '_create_test_web_form_source'

    def _create_test_web_form_source(self, extra_data=None):
        backend_data = {'uncompress': SOURCE_UNCOMPRESS_CHOICE_NEVER}

        if extra_data:
            backend_data.update(extra_data)

        self._create_test_source(
            backend_path=SourceBackendWebForm.get_class_path(),
            backend_data=backend_data
        )



import logging

from django.apps import apps
from django.core.exceptions import ImproperlyConfigured
from django.db import models
from django.utils.encoding import force_text
from django.utils.functional import classproperty
from django.utils.translation import ugettext_lazy as _

from mayan.apps.common.class_mixins import AppsModuleLoaderMixin
from mayan.apps.common.utils import get_class_full_name

logger = logging.getLogger(name=__name__)


class BackendMetaclass(type):
    def __new__(mcs, name, bases, attrs):
        new_class = super().__new__(
            mcs, name, bases, attrs
        )

        base_backend_class = BackendMetaclass._get_base_backend_class(
            klass=new_class
        )

        _loader_module_name = getattr(new_class, '_loader_module_name')

        # Check if `_loader_module_name` is set as to not process the
        # metaclass.
        if _loader_module_name and base_backend_class:
            if base_backend_class != new_class:
                # Get this new child class full name, to be used as the
                # registry key.
                new_class_full_name = get_class_full_name(klass=new_class)
                if _loader_module_name in new_class_full_name:
                    # Only load classes defined in the loader module.
                    # This ensures miscellaneous classes like NullBackends
                    # are not registered.
                    base_backend_class.register(klass=new_class)

        return new_class

    @staticmethod
    def _get_base_backend_class(klass):
        # Filter class hierarchy. Select only the class with App loader
        # support. This means that is the class meant to work as the
        # app's backend class.
        base_backend_class = [
            parent for parent in klass.mro() if '_loader_module_name' in parent.__dict__ and parent._loader_module_name
        ]

        # No more than one app backend class should be returned.
        if len(base_backend_class) > 1:
            raise ImproperlyConfigured(
                'More than one backend parent class was returned. Should be '
                'only one.'
            )

        if base_backend_class:
            return base_backend_class[0]
        else:
            return None


class BaseBackend(AppsModuleLoaderMixin, metaclass=BackendMetaclass):
    _backend_identifier = 'backend_class_path'
    _registry = {}

    @classproperty
    def backend_class_path(cls):
        return cls.get_class_path()

    @classproperty
    def backend_id(cls):
        return getattr(cls, cls._backend_identifier)

    @classmethod
    def get(cls, name):
        return cls._registry.get(cls, {})[name]

    @classmethod
    def get_all(cls):
        return list(cls._registry.get(cls, {}).values())

    @classmethod
    def get_choices(cls):
        choices = [
            (
                backend.backend_id, backend.label
            ) for backend in cls.get_all()
        ]

        choices.sort(key=lambda x: x[1])

        return choices

    @classmethod
    def get_class_path(cls):
        return get_class_full_name(klass=cls)

    @classmethod
    def register(cls, klass):
        # Initialize the app backend class registry.
        cls._registry.setdefault(
            cls, {}
        )

        # Get this new child class full name, to be used as the
        # registry key.
        registry_key = klass.backend_id

        # Add the new child class to the app backend class registry.
        cls._registry[cls][registry_key] = klass


class ModelAttribute:
    _class_registry = []
    _model_registry = {}

    @classmethod
    def get_all_choices_for(cls, model):
        result = []

        for klass in cls._class_registry:
            klass_choices = klass.get_choices_for(model=model)
            if klass_choices:
                result.append(
                    (klass.class_label, klass_choices)
                )

        return result

    @classmethod
    def get_choices_for(cls, model):
        return sorted(
            (
                (
                    entry.name, entry.get_display()
                ) for entry in cls.get_for(model=model)
            ), key=lambda x: x[1]
        )

    @classmethod
    def get_for(cls, model):
        try:
            return cls._model_registry[cls.class_name][model]
        except KeyError:
            # We were passed a model instance, try again using the model of
            # the instance.

            # If we are already in the model class, exit with an error.
            if model.__class__ == models.base.ModelBase:
                return []

            return cls.get_for(model=type(model))

    @classmethod
    def register(cls, klass):
        cls._class_registry.append(klass)

    def __init__(self, model, name, label=None, description=None):
        self.model = model
        self.label = label
        self.name = name
        self.description = description
        self._model_registry.setdefault(
            self.class_name, {}
        )
        self._model_registry[self.class_name].setdefault(
            model, []
        )
        self._model_registry[self.class_name][model].append(self)

    def get_display(self, show_name=False):
        if self.description:
            return '{} - {}'.format(
                self.name if show_name else self.label, self.description
            )
        else:
            return force_text(s=self.name if show_name else self.label)


class ModelBaseBackend(BaseBackend):
    _backend_app_label = None
    _backend_model_name = None

    @classmethod
    def register(cls, klass):
        super().register(klass=klass)

    def get_model_instance(self):
        BackendModel = apps.get_model(
            app_label=self._backend_app_label,
            model_name=self._backend_model_name
        )

        return BackendModel.objects.get(pk=self.model_instance_id)

    def __init__(self, model_instance_id, **kwargs):
        self.model_instance_id = model_instance_id
        self.kwargs = kwargs


class ModelField(ModelAttribute):
    class_label = _('Model fields')
    class_name = 'field'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._final_model_verbose_name = None

        if not self.label:
            self.label = self.get_field_attribute(
                attribute='verbose_name'
            )
            if self.label != self._final_model_verbose_name:
                self.label = '{}, {}'.format(
                    self._final_model_verbose_name, self.label
                )

        if not self.description:
            self.description = self.get_field_attribute(
                attribute='help_text'
            )

    def get_field_attribute(self, attribute, model=None, field_name=None):
        if not model:
            model = self.model

        if not field_name:
            field_name = self.name

        parts = field_name.split('__')
        if len(parts) > 1:
            return self.get_field_attribute(
                attribute=attribute,
                field_name='__'.join(
                    parts[1:]
                ),
                model=model._meta.get_field(
                    field_name=parts[0]
                ).related_model
            )
        else:
            self._final_model_verbose_name = model._meta.verbose_name
            return getattr(
                model._meta.get_field(field_name=field_name),
                attribute
            )


class ModelFieldRelated(ModelField):
    class_label = _('Model related fields')
    class_name = 'related_field'


class ModelProperty(ModelAttribute):
    class_label = _('Model properties')
    class_name = 'property'


class ModelReverseField(ModelField):
    class_label = _('Model reverse fields')
    class_name = 'reverse_field'

    def __init__(self, *args, **kwargs):
        super(ModelField, self).__init__(*args, **kwargs)
        self._final_model_verbose_name = None

        if not self.label:
            self.label = self.get_field_attribute(
                attribute='verbose_name_plural'
            )

    def get_field_attribute(self, attribute, model=None, field_name=None):
        if not model:
            model = self.model

        if not field_name:
            field_name = self.name

        return getattr(
            model._meta.get_field(field_name=field_name).related_model._meta,
            attribute
        )


class ModelQueryFields:
    _registry = {}

    @classmethod
    def get(cls, model):
        try:
            return cls._registry[model]
        except KeyError:
            ModelQueryFields(model=model)
            return cls.get(model=model)

    def __init__(self, model):
        self.model = model
        self.select_related_fields = []
        self.prefetch_related_fields = []
        self.__class__._registry[model] = self

    def add_select_related_field(self, field_name):
        if field_name in self.select_related_fields:
            raise ImproperlyConfigured(
                '"{}" model already has a "{}" query select related field.'.format(
                    self.model, field_name
                )
            )
        self.select_related_fields.append(field_name)

    def add_prefetch_related_field(self, field_name):
        if field_name in self.prefetch_related_fields:
            raise ImproperlyConfigured(
                '"{}" model already has a "{}" query prefetch related field.'.format(
                    self.model, field_name
                )
            )
        self.prefetch_related_fields.append(field_name)

    def get_queryset(self, manager_name=None):
        if manager_name:
            manager = getattr(self.model, manager_name)
        else:
            manager = self.model._meta.default_manager

        queryset = manager.all()

        if self.select_related_fields:
            queryset = queryset.select_related(*self.select_related_fields)

        if self.prefetch_related_fields:
            queryset = queryset.prefetch_related(
                *self.prefetch_related_fields
            )

        return queryset


class QuerysetParametersSerializer:
    @staticmethod
    def decompose(_model, _method_name, _manager_name=None, **kwargs):
        ContentType = apps.get_model(
            app_label='contenttypes', model_name='ContentType'
        )

        _manager_name = _manager_name or _model._meta.default_manager.name

        serialized_kwargs = []
        for name, value in kwargs.items():
            try:
                content_type = ContentType.objects.get_for_model(model=value)
            except AttributeError:
                """The value is not a model instance, pass it as-is."""
                serialized_kwargs.append(
                    {
                        'name': name,
                        'value': value
                    }
                )
            else:
                serialized_kwargs.append(
                    {
                        'name': name,
                        'content_type_id': content_type.pk,
                        'object_id': value.pk
                    }
                )

        return {
            'model_content_type_id': ContentType.objects.get_for_model(
                model=_model
            ).pk,
            'manager_name': _manager_name,
            'method_name': _method_name,
            'kwargs': serialized_kwargs
        }

    @staticmethod
    def rebuild(decomposed_queryset):
        ContentType = apps.get_model(
            app_label='contenttypes', model_name='ContentType'
        )

        model = ContentType.objects.get(
            pk=decomposed_queryset['model_content_type_id']
        ).model_class()

        queryset = getattr(
            model, decomposed_queryset['manager_name']
        )

        kwargs = {}

        for parameter in decomposed_queryset.get('kwargs', ()):
            if 'content_type_id' in parameter:
                content_type = ContentType.objects.get(
                    pk=parameter['content_type_id']
                )
                value = content_type.get_object_for_this_type(
                    pk=parameter['object_id']
                )
            else:
                value = parameter['value']

            kwargs[
                parameter['name']
            ] = value

        return getattr(
            queryset, decomposed_queryset['method_name']
        )(**kwargs)


class StoredBaseBackend(BaseBackend):
    _backend_app_label = None
    _backend_model_name = None

    @classmethod
    def register(cls, klass):
        super().register(klass=klass)
        klass.get_model_instance()

    @classmethod
    def get_model_instance(cls):
        StoredBackendModel = apps.get_model(
            app_label=cls._backend_app_label,
            model_name=cls._backend_model_name
        )

        backend_path = get_class_full_name(klass=cls)

        instance, created = StoredBackendModel.objects.get_or_create(
            backend_path=backend_path
        )
        cls._backend_stored_model_id = instance.pk
        return instance

    def __init__(self, **kwargs):
        self.kwargs = kwargs


ModelAttribute.register(klass=ModelProperty)
ModelAttribute.register(klass=ModelField)
ModelAttribute.register(klass=ModelFieldRelated)
ModelAttribute.register(klass=ModelReverseField)



from django.db import models
from django.utils.translation import ugettext_lazy as _

from mayan.apps.databases.model_mixins import ExtraDataModelMixin
from mayan.apps.documents.models.document_type_models import DocumentType
from mayan.apps.documents.models.document_version_page_models import DocumentVersionPage
from mayan.apps.events.classes import EventManagerSave
from mayan.apps.events.decorators import method_event

from .events import event_ocr_document_version_page_content_edited
from .managers import (
    DocumentVersionPageOCRContentManager, DocumentTypeSettingsManager
)


class DocumentTypeOCRSettings(ExtraDataModelMixin, models.Model):
    """
    Model to store the OCR settings for a document type.
    """
    document_type = models.OneToOneField(
        on_delete=models.CASCADE, related_name='ocr_settings',
        to=DocumentType, unique=True, verbose_name=_('Document type')
    )
    auto_ocr = models.BooleanField(
        default=True, help_text=_(
            'Automatically queue newly created documents for OCR.'
        ), verbose_name=_('Auto OCR')
    )

    objects = DocumentTypeSettingsManager()

    class Meta:
        verbose_name = _('Document type settings')
        verbose_name_plural = _('Document types settings')

    def natural_key(self):
        return self.document_type.natural_key()
    natural_key.dependencies = ['documents.DocumentType']


class DocumentVersionPageOCRContent(models.Model):
    """
    This model stores the OCR results for a document version page.
    """
    document_version_page = models.OneToOneField(
        on_delete=models.CASCADE, related_name='ocr_content',
        to=DocumentVersionPage, verbose_name=_('Document version page')
    )
    content = models.TextField(
        blank=True, help_text=_(
            'The actual text content extracted by the OCR backend.'
        ), verbose_name=_('Content')
    )

    objects = DocumentVersionPageOCRContentManager()

    class Meta:
        verbose_name = _('Document version page OCR content')
        verbose_name_plural = _('Document version pages OCR contents')

    def __str__(self):
        return str(self.document_version_page)

    @method_event(
        event_manager_class=EventManagerSave,
        edited={
            'action_object': 'document_version_page.document_version.document',
            'event': event_ocr_document_version_page_content_edited,
            'target': 'document_version_page'
        }
    )
    def save(self, *args, **kwargs):
        return super().save(*args, **kwargs)



from django.apps import apps
from django.contrib.contenttypes.models import ContentType

from ..classes import ErrorLog

from .literals import TEST_ERROR_LOG_ENTRY_RESULT


class ErrorLogPartitionEntryAPIViewTestMixin:
    def _request_test_error_log_delete_api_view(self):
        content_type = ContentType.objects.get_for_model(
            model=self._test_object
        )

        return self.delete(
            viewname='rest_api:errorlogpartitionentry-detail', kwargs={
                'app_label': content_type.app_label,
                'model_name': content_type.model,
                'object_id': self._test_object.pk,
                'error_log_partition_entry_id': self._test_error_log_entry.pk
            }
        )

    def _request_test_error_log_detail_api_view(self):
        content_type = ContentType.objects.get_for_model(
            model=self._test_object
        )

        return self.get(
            viewname='rest_api:errorlogpartitionentry-detail', kwargs={
                'app_label': content_type.app_label,
                'model_name': content_type.model,
                'object_id': self._test_object.pk,
                'error_log_partition_entry_id': self._test_error_log_entry.pk
            }
        )

    def _request_test_error_log_list_api_view(self):
        content_type = ContentType.objects.get_for_model(
            model=self._test_object
        )

        return self.get(
            viewname='rest_api:errorlogpartitionentry-list', kwargs={
                'app_label': content_type.app_label,
                'model_name': content_type.model,
                'object_id': self._test_object.pk
            }
        )


class ErrorLogPartitionEntryTestMixin:
    def _create_test_error_log_object(self):
        app_config = apps.get_app_config(app_label='logging')
        self.error_log = ErrorLog(app_config=app_config)

        self._create_test_user()
        self._test_object = self._test_user

    def _create_test_error_log_entry(self):
        self._test_error_log_entry = self._test_object.error_log.create(
            text=TEST_ERROR_LOG_ENTRY_RESULT
        )


class ErrorLogViewTestMixin:
    def _request_object_error_log_list_view(self):
        content_type = ContentType.objects.get_for_model(
            model=self._test_object
        )

        return self.get(
            viewname='logging:object_error_log_entry_list', kwargs={
                'app_label': content_type.app_label,
                'model_name': content_type.model,
                'object_id': self._test_object.pk
            }
        )

    def _request_object_error_log_clear_view(self):
        content_type = ContentType.objects.get_for_model(
            model=self._test_object
        )

        return self.post(
            viewname='logging:object_error_log_entry_list_clear', kwargs={
                'app_label': content_type.app_label,
                'model_name': content_type.model,
                'object_id': self._test_object.pk
            }
        )

    def _request_object_error_log_entry_delete_view(self):
        content_type = ContentType.objects.get_for_model(
            model=self._test_object
        )

        return self.post(
            viewname='logging:object_error_log_entry_delete', kwargs={
                'app_label': content_type.app_label,
                'model_name': content_type.model,
                'object_id': self._test_object.pk,
                'error_log_partition_entry_id': self._test_error_log_entry.pk
            }
        )


class GlobalErrorLogViewTestMixin:
    def _request_global_error_log_partition_entry_list_view(self):
        return self.get(
            viewname='logging:global_error_log_partition_entry_list'
        )



from django.utils.translation import ugettext_lazy as _

from rest_framework.reverse import reverse
from rest_framework_recursive.fields import RecursiveField

from mayan.apps.documents.models.document_type_models import DocumentType
from mayan.apps.documents.permissions import permission_document_type_edit
from mayan.apps.rest_api import serializers
from mayan.apps.rest_api.relations import FilteredPrimaryKeyRelatedField

from .models import (
    IndexInstance, IndexInstanceNode, IndexTemplate, IndexTemplateNode
)


class IndexInstanceSerializer(serializers.ModelSerializer):
    url = serializers.SerializerMethodField(
        label=_('URL'), read_only=True
    )
    nodes_url = serializers.SerializerMethodField(
        label=_('Nodes URL'), read_only=True
    )

    class Meta:
        fields = ('label', 'id', 'nodes_url', 'url')
        model = IndexInstance
        read_only_fields = fields

    def get_url(self, obj):
        return reverse(
            viewname='rest_api:indexinstance-detail', kwargs={
                'index_instance_id': obj.pk,
            }, format=self.context['format'], request=self.context['request']
        )

    def get_nodes_url(self, obj):
        return reverse(
            viewname='rest_api:indexinstancenode-list', kwargs={
                'index_instance_id': obj.pk,
            }, format=self.context['format'], request=self.context['request']
        )


class IndexInstanceNodeSerializer(serializers.ModelSerializer):
    children_url = serializers.SerializerMethodField(
        label=_('Children URL'), read_only=True
    )
    documents_url = serializers.SerializerMethodField(
        label=_('Documents URL'), read_only=True
    )
    index_url = serializers.SerializerMethodField(
        label=_('Index URL'), read_only=True
    )
    parent_url = serializers.SerializerMethodField(
        label=_('Parent URL'), read_only=True
    )
    url = serializers.SerializerMethodField(
        label=_('URL'), read_only=True
    )

    class Meta:
        fields = (
            'documents_url', 'children_url', 'id', 'index_url', 'level',
            'parent_id', 'parent_url', 'value', 'url'
        )
        model = IndexInstanceNode
        read_only_fields = fields

    def get_children_url(self, obj):
        return reverse(
            viewname='rest_api:indexinstancenode-children-list', kwargs={
                'index_instance_id': obj.index().pk,
                'index_instance_node_id': obj.pk
            }, format=self.context['format'], request=self.context['request']
        )

    def get_documents_url(self, obj):
        return reverse(
            viewname='rest_api:indexinstancenode-document-list', kwargs={
                'index_instance_id': obj.index().pk,
                'index_instance_node_id': obj.pk
            }, format=self.context['format'], request=self.context['request']
        )

    def get_index_url(self, obj):
        return reverse(
            viewname='rest_api:indexinstance-detail', kwargs={
                'index_instance_id': obj.index().pk,
            }, format=self.context['format'], request=self.context['request']
        )

    def get_parent_url(self, obj):
        if obj.parent and not obj.parent.is_root_node():
            return reverse(
                viewname='rest_api:indexinstancenode-detail', kwargs={
                    'index_instance_id': obj.index().pk,
                    'index_instance_node_id': obj.parent.pk
                }, format=self.context['format'],
                request=self.context['request']
            )
        else:
            return ''

    def get_url(self, obj):
        return reverse(
            viewname='rest_api:indexinstancenode-detail', kwargs={
                'index_instance_id': obj.index().pk,
                'index_instance_node_id': obj.pk
            }, format=self.context['format'], request=self.context['request']
        )


class IndexTemplateNodeSerializer(serializers.ModelSerializer):
    children = RecursiveField(
        label=_('Children'), many=True, read_only=True
    )
    index_url = serializers.SerializerMethodField(
        label=_('Index URL'), read_only=True
    )
    parent_url = serializers.SerializerMethodField(
        label=_('Parent URL'), read_only=True
    )
    url = serializers.SerializerMethodField(
        label=_('URL'), read_only=True
    )

    # DEPRECATION: Version 5.0, remove 'parent' from GET fields as this
    # is replaced by 'parent_id'.
    # DEPRECATION: Version 5.0, remove 'index' from GET fields as this
    # is replaced by 'index_id'.
    class Meta:
        fields = (
            'children', 'enabled', 'expression', 'id', 'index', 'index_id',
            'index_url', 'level', 'link_documents', 'parent', 'parent_id',
            'parent_url', 'url'
        )
        model = IndexTemplateNode
        read_only_fields = (
            'children', 'id', 'index', 'index_id', 'index_url', 'level',
            'parent_id', 'parent_url', 'url'
        )

    def get_index_url(self, obj):
        return reverse(
            viewname='rest_api:indextemplate-detail', kwargs={
                'index_template_id': obj.index.pk,
            }, format=self.context['format'], request=self.context['request']
        )

    def get_parent_url(self, obj):
        if obj.parent and not obj.parent.is_root_node():
            return reverse(
                viewname='rest_api:indextemplatenode-detail', kwargs={
                    'index_template_id': obj.index.pk,
                    'index_template_node_id': obj.parent.pk
                }, format=self.context['format'],
                request=self.context['request']
            )
        else:
            return ''

    def get_url(self, obj):
        return reverse(
            viewname='rest_api:indextemplatenode-detail', kwargs={
                'index_template_id': obj.index.pk,
                'index_template_node_id': obj.pk
            }, format=self.context['format'], request=self.context['request']
        )


class IndexTemplateNodeWriteSerializer(serializers.ModelSerializer):
    children = RecursiveField(
        label=_('Children'), many=True, read_only=True
    )
    index_url = serializers.SerializerMethodField(
        label=_('Index URL'), read_only=True
    )
    parent = FilteredPrimaryKeyRelatedField(
        label=_('Parent')
    )
    parent_url = serializers.SerializerMethodField(
        label=_('Parent URL'), read_only=True
    )
    url = serializers.SerializerMethodField(
        label=_('URL'), read_only=True
    )

    class Meta:
        fields = (
            'children', 'enabled', 'expression', 'id', 'index', 'index_url',
            'level', 'link_documents', 'parent', 'parent_url', 'url'
        )
        model = IndexTemplateNode
        read_only_fields = (
            'children', 'id', 'index', 'index_url', 'level', 'parent_url',
            'url'
        )

    def create(self, validated_data):
        validated_data['index'] = self.context['index_template']
        return super().create(validated_data=validated_data)

    def get_index_url(self, obj):
        return reverse(
            viewname='rest_api:indextemplate-detail', kwargs={
                'index_template_id': obj.index.pk,
            }, format=self.context['format'], request=self.context['request']
        )

    def get_parent_queryset(self):
        return self.context['index_template'].index_template_nodes.all()

    def get_parent_url(self, obj):
        if obj.parent:
            return reverse(
                viewname='rest_api:indextemplatenode-detail', kwargs={
                    'index_template_id': obj.index.pk,
                    'index_template_node_id': obj.parent.pk
                }, format=self.context['format'],
                request=self.context['request']
            )
        else:
            return ''

    def get_url(self, obj):
        return reverse(
            viewname='rest_api:indextemplatenode-detail', kwargs={
                'index_template_id': obj.index.pk,
                'index_template_node_id': obj.pk
            }, format=self.context['format'], request=self.context['request']
        )


class IndexTemplateSerializer(serializers.HyperlinkedModelSerializer):
    document_types_url = serializers.HyperlinkedIdentityField(
        help_text=_(
            'URL of the API endpoint showing the list document types '
            'associated with this index template.'
        ), label=_('Document types URL'),
        lookup_url_kwarg='index_template_id',
        view_name='rest_api:indextemplate-documenttype-list'
    )
    document_types_add_url = serializers.HyperlinkedIdentityField(
        help_text=_(
            'URL of the API endpoint to add document types '
            'to this index template.'
        ), label=_('Document types add URL'),
        lookup_url_kwarg='index_template_id',
        view_name='rest_api:indextemplate-documenttype-add'
    )
    document_types_remove_url = serializers.HyperlinkedIdentityField(
        help_text=_(
            'URL of the API endpoint to remove document types '
            'from this index template.'
        ), label=_('Document types remove URL'),
        lookup_url_kwarg='index_template_id',
        view_name='rest_api:indextemplate-documenttype-remove'
    )
    index_template_root_node_id = serializers.PrimaryKeyRelatedField(
        label=_('Index template root node ID'),
        source='index_template_root_node', read_only=True
    )
    nodes_url = serializers.SerializerMethodField(
        label=_('Nodes URL'), read_only=True
    )
    rebuild_url = serializers.HyperlinkedIdentityField(
        label=_('Rebuild URL'), lookup_url_kwarg='index_template_id',
        view_name='rest_api:indextemplate-rebuild',
    )
    reset_url = serializers.HyperlinkedIdentityField(
        label=_('Reset URL'), lookup_url_kwarg='index_template_id',
        view_name='rest_api:indextemplate-reset',
    )
    url = serializers.SerializerMethodField(
        label=_('URL'), read_only=True
    )

    class Meta:
        extra_kwargs = {
            'document_types': {
                'label': _('Document types'),
                'lookup_url_kwarg': 'document_type_id',
                'view_name': 'rest_api:documenttype-detail'
            },
        }
        fields = (
            'document_types_add_url', 'document_types_url',
            'document_types_remove_url', 'enabled', 'id',
            'index_template_root_node_id', 'label', 'nodes_url',
            'rebuild_url', 'reset_url', 'slug', 'url'
        )
        model = IndexTemplate
        read_only_fields = (
            'document_types_add_url', 'document_types_url',
            'document_types_remove_url', 'id', 'index_template_root_node_id',
            'nodes_url', 'rebuild_url', 'reset_url', 'url'
        )

    def get_url(self, obj):
        return reverse(
            viewname='rest_api:indextemplate-detail', kwargs={
                'index_template_id': obj.pk,
            }, format=self.context['format'], request=self.context['request']
        )

    def get_nodes_url(self, obj):
        return reverse(
            viewname='rest_api:indextemplatenode-list', kwargs={
                'index_template_id': obj.pk,
            }, format=self.context['format'], request=self.context['request']
        )


class DocumentTypeAddSerializer(serializers.Serializer):
    document_type = FilteredPrimaryKeyRelatedField(
        help_text=_(
            'Primary key of the document type to add to the index template.'
        ), label=_('Document type ID'), source_model=DocumentType,
        source_permission=permission_document_type_edit
    )


class DocumentTypeRemoveSerializer(serializers.Serializer):
    document_type = FilteredPrimaryKeyRelatedField(
        help_text=_(
            'Primary key of the document type to remove from the index template.'
        ), label=_('Document type ID'), source_model=DocumentType,
        source_permission=permission_document_type_edit
    )



__title__ = 'Mayan EDMS'
__version__ = '4.4'
__build__ = 0x040301
__build_string__ = 'v4.4-17-g34bf9cc670_Tue Jan 17 04:00:08 2023 -0500'
__django_version__ = '3.2'
__author__ = 'Roberto Rosario'
__author_email__ = 'roberto.rosario@mayan-edms.com'
__description__ = 'Free Open Source Electronic Document Management System'
__license__ = 'Apache 2.0'
__copyright_short__ = '2011 Roberto Rosario'
__copyright__ = '{} {}'.format('Copyright', __copyright_short__)
__website__ = 'https://www.mayan-edms.com'



import base64
import logging
import os
from pathlib import Path
import re
import time
from urllib.parse import quote_plus, unquote_plus

from furl import furl

from django import forms
from django.conf import settings
from django.contrib import messages
from django.core.files import File
from django.core.files.base import ContentFile
from django.http import Http404, StreamingHttpResponse
from django.utils.encoding import force_text
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _
from django.views.generic.list import MultipleObjectMixin

from rest_framework import status
from rest_framework.generics import get_object_or_404 as rest_get_object_or_404
from rest_framework.response import Response
from rest_framework.reverse import reverse as rest_framework_reverse

from mayan.apps.acls.models import AccessControlList
from mayan.apps.appearance.classes import Icon
from mayan.apps.common.menus import menu_object
from mayan.apps.converter.classes import ConverterBase
from mayan.apps.converter.exceptions import InvalidOfficeFormat
from mayan.apps.converter.transformations import TransformationResize
from mayan.apps.converter.utils import IndexedDictionary
from mayan.apps.documents.html_widgets import ThumbnailWidget
from mayan.apps.documents.models.document_type_models import DocumentType
from mayan.apps.documents.permissions import permission_document_create
from mayan.apps.navigation.classes import Link, SourceColumn
from mayan.apps.sources.classes import SourceBackend
from mayan.apps.sources.forms import UploadBaseForm
from mayan.apps.sources.literals import STORAGE_NAME_SOURCE_CACHE_FOLDER
from mayan.apps.sources.source_backends.mixins import (
    SourceBackendCompressedMixin, SourceBackendInteractiveMixin,
    SourceBaseMixin
)
from mayan.apps.storage.classes import DefinedStorage
from mayan.apps.storage.models import SharedUploadedFile
from mayan.apps.views.settings import setting_paginate_by

from ..classes import SourceBackendAction
from ..tasks import task_process_document_upload

from .literals import (
    REGULAR_EXPRESSION_MATCH_EVERYTHING, REGULAR_EXPRESSION_MATCH_NOTHING
)

__all__ = ('SourceBackendStagingFolder',)
logger = logging.getLogger(name=__name__)


class StagingFolderFile:
    """
    Simple class to extend the File class to add preview capabilities
    files in a directory on a storage.
    """
    def __init__(self, staging_folder, filename=None, encoded_filename=None):
        self.staging_folder = staging_folder
        if encoded_filename:
            self.encoded_filename = str(encoded_filename)

            try:
                self.filename = base64.urlsafe_b64decode(
                    s=unquote_plus(self.encoded_filename)
                ).decode('utf8')
            except UnicodeDecodeError:
                raise ValueError(
                    'Incorrect `encoded_filename` value.'
                )
        else:
            if not filename:
                raise KeyError(
                    'Supply either `encoded_filename` or `filename` when '
                    'instantiating a staging folder file.'
                )
            self.filename = filename
            self.encoded_filename = quote_plus(
                base64.urlsafe_b64encode(
                    s=filename.encode('utf8')
                )
            )

    def __str__(self):
        return force_text(s=self.filename)

    def as_file(self):
        return File(
            file=open(
                file=self.get_full_path(), mode='rb'
            ), name=self.filename
        )

    @property
    def cache_filename(self):
        return '{}-{}'.format(
            self.staging_folder.model_instance_id, self.encoded_filename
        )

    def delete(self):
        self.storage.delete(self.cache_filename)
        os.unlink(
            self.get_full_path()
        )

    def generate_image(self, transformation_instance_list=None):
        # Check is transformed image is available.
        logger.debug(
            'transformations cache filename: %s', self.cache_filename
        )

        if self.storage.exists(self.cache_filename):
            logger.debug(
                'staging file cache file "%s" found', self.cache_filename
            )
        else:
            logger.debug(
                'staging file cache file "%s" not found', self.cache_filename
            )
            image = self.get_image(
                transformation_instance_list=transformation_instance_list
            )

            # Since open "wb+" doesn't create files, check if the file
            # exists, if not then create it.
            self.storage.save(
                content=ContentFile(content=b''), name=self.cache_filename
            )

            with self.storage.open(name=self.cache_filename, mode='wb+') as file_object:
                file_object.write(
                    image.getvalue()
                )

        return self.cache_filename

    def get_api_image_url(
        self, maximum_layer_order=None, transformation_instance_list=None,
        request=None, user=None
    ):
        final_url = furl()
        final_url.args = {'encoded_filename': self.encoded_filename}
        final_url.path = rest_framework_reverse(
            'rest_api:source-action', kwargs={
                'action_name': 'file_image',
                'source_id': self.staging_folder.model_instance_id
            }, request=request
        )

        return final_url.tostr()

    def get_combined_transformation_list(
        self, maximum_layer_order=None, transformation_instance_list=None,
        user=None
    ):
        """
        Return a list of transformation containing the server side
        transformations for this object as well as transformations
        created from the arguments as transient interactive transformation.
        """
        result = [
            TransformationResize(
                width=self.staging_folder.kwargs['preview_width'],
                height=self.staging_folder.kwargs['preview_height'],
            )
        ]

        # Interactive transformations second.
        result.extend(transformation_instance_list or [])

        return result

    def get_date_time_created(self):
        return time.ctime(
            os.path.getctime(
                self.get_full_path()
            )
        )

    def get_full_path(self):
        return os.path.join(
            self.staging_folder.kwargs['folder_path'], self.filename
        )

    def get_image(self, transformation_instance_list=None):
        try:
            with open(file=self.get_full_path(), mode='rb') as file_object:
                converter = ConverterBase.get_converter_class()(
                    file_object=file_object
                )

                try:
                    with converter.to_pdf() as pdf_file_object:
                        image_converter = ConverterBase.get_converter_class()(
                            file_object=pdf_file_object
                        )
                        page_image = image_converter.get_page()
                except InvalidOfficeFormat:
                    page_image = converter.get_page()
        except Exception as exception:
            # Cleanup in case of error.
            logger.error(
                'Error getting staging file image for file "%s"; %s',
                self.get_full_path(), exception
            )
            raise
        else:
            return page_image

    @cached_property
    def storage(self):
        return DefinedStorage.get(
            name=STORAGE_NAME_SOURCE_CACHE_FOLDER
        ).get_storage_instance()


class StagingUploadForm(UploadBaseForm):
    """
    Form that show all the files in the staging folder specified by the
    StagingFolderFile class passed as 'cls' argument.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        try:
            self.fields['staging_folder_file_id'].choices = [
                (
                    staging_folder_file.encoded_filename, str(
                        staging_folder_file
                    )
                ) for staging_folder_file in self.source.get_backend_instance().get_files()
            ]
        except Exception as exception:
            logger.error('exception: %s', exception)

    staging_folder_file_id = forms.ChoiceField(
        label=_('Staging file'), widget=forms.widgets.Select(
            attrs={'class': 'select2'}
        )
    )


class SourceBackendStagingFolder(
    SourceBackendCompressedMixin, SourceBackendInteractiveMixin,
    SourceBaseMixin, SourceBackend
):
    actions = (
        SourceBackendAction(
            name='file_delete', arguments=('encoded_filename',)
        ),
        SourceBackendAction(
            name='file_image', arguments=('encoded_filename',),
            confirmation=False
        ),
        SourceBackendAction(name='file_list', confirmation=False),
        SourceBackendAction(
            name='file_upload', arguments=(
                'document_type_id', 'encoded_filename', 'expand'
            )
        )
    )
    field_order = (
        'folder_path', 'preview_width', 'preview_height',
        'delete_after_upload', 'include_subdirectories', 'include_regex',
        'exclude_regex'
    )
    fields = {
        'folder_path': {
            'class': 'django.forms.CharField',
            'default': '',
            'help_text': _(
                'Server side filesystem path.'
            ),
            'kwargs': {
                'max_length': 255,
            },
            'label': _('Folder path'),
            'required': True
        },
        'preview_width': {
            'class': 'django.forms.IntegerField',
            'help_text': _(
                'Width value to be passed to the converter backend.'
            ),
            'kwargs': {
                'min_value': 0
            },
            'label': _('Preview width'),
            'required': True
        },
        'preview_height': {
            'class': 'django.forms.IntegerField',
            'help_text': _(
                'Height value to be passed to the converter backend.'
            ),
            'kwargs': {
                'min_value': 0
            },
            'label': _('Preview height'),
            'required': False
        },
        'delete_after_upload': {
            'class': 'django.forms.BooleanField',
            'help_text': _(
                'Delete the file after is has been successfully uploaded.'
            ),
            'label': _('Delete after upload'),
            'required': False
        },
        'include_subdirectories': {
            'class': 'django.forms.BooleanField',
            'default': '',
            'help_text': _(
                'If checked, not only will the folder path be scanned for '
                'files but also its subdirectories.'
            ),
            'label': _('Include subdirectories?'),
            'required': False
        },
        'include_regex': {
            'class': 'django.forms.CharField',
            'default': '',
            'help_text': _(
                'Regular expression used to select which files to upload.'
            ),
            'label': _('Include regular expression'),
            'required': False
        },
        'exclude_regex': {
            'class': 'django.forms.CharField',
            'default': '',
            'help_text': _(
                'Regular expression used to exclude which files to upload.'
            ),
            'label': _('Exclude regular expression'),
            'required': False
        }
    }
    icon_staging_folder_file = Icon(driver_name='fontawesome', symbol='file')
    icon_staging_folder_file_delete = Icon(
        driver_name='fontawesome', symbol='times'
    )
    icon_staging_folder_file_select = Icon(
        driver_name='fontawesome', symbol='check'
    )
    label = _('Staging folder')
    upload_form_class = StagingUploadForm

    @classmethod
    def intialize(cls):
        link_staging_folder_file_delete = Link(
            icon=cls.icon_staging_folder_file_delete, kwargs={
                'source_id': 'source.pk',
                'action_name': '"file_delete"'
            }, permissions=(permission_document_create,), query={
                'document_id': 'document.pk',
                'document_type_id': 'document_type.pk',
                'source_id': 'source.pk',
                'encoded_filename': 'object.encoded_filename'
            }, tags='dangerous', text=_('Delete'),
            view='sources:source_action'
        )
        link_staging_folder_file_select = Link(
            html_data={'encoded_filename': 'object.encoded_filename'},
            html_extra_classes='sources-staging_folder-select',
            icon=cls.icon_staging_folder_file_select, text=_('Select'),
            url=''
        )

        class StagingFolderFileThumbnailWidget(ThumbnailWidget):
            gallery_name = 'sources:staging_list'

            def disable_condition(self, instance):
                return True

        SourceColumn(
            func=lambda context: context['object'].get_date_time_created(),
            label=_('Created'), source=StagingFolderFile
        )

        SourceColumn(
            html_extra_classes='text-center', label=_('Thumbnail'),
            source=StagingFolderFile,
            widget=StagingFolderFileThumbnailWidget
        )

        menu_object.bind_links(
            links=(
                link_staging_folder_file_delete,
                link_staging_folder_file_select
            ), sources=(StagingFolderFile,)
        )

    def action_file_delete(self, request, encoded_filename):
        staging_folder_file = self.get_file(
            encoded_filename=encoded_filename
        )
        staging_folder_file.delete()

    def action_file_image(self, request, encoded_filename, **kwargs):
        encoded_filename = encoded_filename[0]

        query_dict = request.GET

        transformation_instance_list = IndexedDictionary(
            dictionary=query_dict
        ).as_instance_list()

        maximum_layer_order = request.GET.get('maximum_layer_order')
        if maximum_layer_order:
            maximum_layer_order = int(maximum_layer_order)

        staging_folder_file = self.get_file(
            encoded_filename=encoded_filename
        )

        combined_transformation_list = staging_folder_file.get_combined_transformation_list(
            maximum_layer_order=maximum_layer_order,
            transformation_instance_list=transformation_instance_list,
            user=request.user
        )

        cache_filename = staging_folder_file.generate_image(
            transformation_instance_list=combined_transformation_list
        )

        storage_staging_folder_file_image_cache = DefinedStorage.get(
            name=STORAGE_NAME_SOURCE_CACHE_FOLDER
        ).get_storage_instance()

        def file_generator():
            with storage_staging_folder_file_image_cache.open(name=cache_filename) as file_object:
                converter = ConverterBase.get_converter_class()(
                    file_object=file_object
                )
                for transformation in combined_transformation_list or ():
                    converter.transform(transformation=transformation)

                result = converter.get_page()

                while True:
                    chunk = result.read(File.DEFAULT_CHUNK_SIZE)
                    if not chunk:
                        break
                    else:
                        yield chunk

        response = StreamingHttpResponse(
            content_type='image', streaming_content=file_generator()
        )
        return None, response

    def action_file_list(self, request):
        staging_folder_files = []

        for staging_folder_file in self.get_files():
            staging_folder_files.append(
                {
                    'filename': staging_folder_file.filename,
                    'delete-url': rest_framework_reverse(
                        viewname='rest_api:source-action', kwargs={
                            'action_name': 'file_delete',
                            'source_id': staging_folder_file.staging_folder.model_instance_id
                        }, request=request
                    ),
                    'encoded_filename': staging_folder_file.encoded_filename,
                    'image-url': staging_folder_file.get_api_image_url(
                        request=request
                    ),
                    'upload-url': rest_framework_reverse(
                        viewname='rest_api:source-action', kwargs={
                            'action_name': 'file_upload',
                            'source_id': staging_folder_file.staging_folder.model_instance_id
                        }, request=request
                    )
                }
            )

        return staging_folder_files, None

    def action_file_upload(
        self, request, document_type_id, encoded_filename, expand=False
    ):
        staging_folder_file = self.get_file(encoded_filename=encoded_filename)

        queryset = AccessControlList.objects.restrict_queryset(
            queryset=DocumentType.objects.all(),
            permission=permission_document_create,
            user=request.user
        )

        document_type = rest_get_object_or_404(
            queryset=queryset, pk=document_type_id
        )

        self.process_kwargs = {
            'request': request,
            'staging_folder_file_filename': staging_folder_file.filename
        }

        shared_uploaded_file = SharedUploadedFile.objects.create(
            file=staging_folder_file.as_file()
        )

        kwargs = {
            'callback_kwargs': self.get_callback_kwargs(),
            'document_type_id': document_type.pk,
            'expand': expand,
            'shared_uploaded_file_id': shared_uploaded_file.pk,
            'source_id': self.model_instance_id,
            'user_id': request.user.pk
        }
        kwargs.update(
            self.get_task_extra_kwargs()
        )

        task_process_document_upload.apply_async(kwargs=kwargs)

        return None, Response(status=status.HTTP_202_ACCEPTED)

    def callback(
        self, document_file, source_id, user_id, extra_data=None,
        query_string=None
    ):
        super().callback(
            document_file=document_file, extra_data=extra_data,
            query_string=query_string, source_id=source_id, user_id=user_id
        )
        if self.kwargs.get('delete_after_upload'):
            path = Path(
                self.kwargs[
                    'folder_path'
                ], extra_data['staging_folder_file_filename']
            )

            try:
                path.unlink()
            except Exception as exception:
                logger.error(
                    'Error deleting staging file: %s; %s',
                    path, exception
                )
                raise Exception(
                    _('Error deleting staging file; %s') % exception
                )

    def get_callback_kwargs(self):
        callback_kwargs = super().get_callback_kwargs()
        callback_kwargs.setdefault('extra_data', {})

        callback_kwargs['extra_data'].update(
            {
                'staging_folder_file_filename': self.process_kwargs[
                    'staging_folder_file_filename'
                ]
            }
        )

        return callback_kwargs

    def get_action_file_delete_context(self, view, encoded_filename):
        staging_folder_file = self.get_file(
            encoded_filename=encoded_filename
        )

        context = {
            'delete_view': True,
            'object': staging_folder_file,
            'object_name': _('Staging file'),
            'title': _('Delete staging file "%s"?') % staging_folder_file
        }

        view_kwargs = view.get_all_kwargs()

        if 'document_type_id' in view_kwargs:
            context['document_type'] = DocumentType.objects.get(
                pk=view_kwargs['document_type_id'][0]
            )

        return context

    def get_file(self, *args, **kwargs):
        try:
            return StagingFolderFile(staging_folder=self, *args, **kwargs)
        except (KeyError, ValueError):
            raise Http404

    def get_files(self):
        include_regex = re.compile(
            pattern=self.kwargs.get(
                'include_regex', REGULAR_EXPRESSION_MATCH_EVERYTHING
            )
        )
        exclude_regex = re.compile(
            pattern=self.kwargs.get(
                'exclude_regex', REGULAR_EXPRESSION_MATCH_NOTHING
            ) or REGULAR_EXPRESSION_MATCH_NOTHING
        )

        path = Path(
            self.kwargs['folder_path']
        )

        # Force path check to trigger any error.
        path.lstat()

        if self.kwargs.get('include_subdirectories', False):
            iterator = path.rglob(pattern='*')
        else:
            iterator = path.glob(pattern='*')

        for entry in sorted(iterator):
            if entry.is_file() and include_regex.match(string=entry.name) and not exclude_regex.match(string=entry.name):
                relative_filename = str(
                    entry.relative_to(
                        self.kwargs['folder_path']
                    )
                )
                yield self.get_file(
                    filename=relative_filename
                )

    def get_shared_uploaded_files(self):
        staging_folder_file = self.get_file(
            encoded_filename=self.process_kwargs['forms']['source_form'].cleaned_data['staging_folder_file_id']
        )
        self.process_kwargs['staging_folder_file_filename'] = staging_folder_file.filename

        return (
            SharedUploadedFile.objects.create(
                file=staging_folder_file.as_file()
            ),
        )

    def get_view_context(self, context, request):
        try:
            staging_files = list(
                self.get_files()
            )
        except Exception as exception:
            messages.error(
                message=_(
                    'Unable get list of staging files; %s'
                ) % exception, request=request
            )
            staging_files = ()
            if settings.DEBUG or settings.TESTING:
                raise

        # Instantiate a fake list view to populate the pagination data for
        # the staging folder file list.
        view = SourceBackendStagingFolderFileListView()
        view.kwargs = self.kwargs
        view.object_list = staging_files
        view.request = request

        template_staging_file_list_context = {
            'hide_link': True,
            'no_results_icon': SourceBackendStagingFolder.icon_staging_folder_file,
            'no_results_text': _(
                'This could mean that the staging folder is '
                'empty. It could also mean that the '
                'operating system user account being used '
                'for Mayan EDMS doesn\'t have the necessary '
                'file system permissions for the folder.'
            ),
            'no_results_title': _(
                'No staging files available'
            )
        }

        template_staging_file_list_context.update(
            view.get_context_data()
        )

        subtemplates_list = [
            {
                'context': {
                    'forms': context['forms']
                },
                'name': 'appearance/generic_multiform_subtemplate.html'
            },
            {
                'context': template_staging_file_list_context,
                'name': 'appearance/generic_list_subtemplate.html'
            }
        ]

        return {
            'subtemplates_list': subtemplates_list
        }


class SourceBackendStagingFolderFileListView(MultipleObjectMixin):
    def get_paginate_by(self, queryset):
        return setting_paginate_by.value



from mayan.apps.appearance.classes import Icon

# App

icon_group = Icon(driver_name='fontawesome', symbol='users')
icon_user_setup = Icon(driver_name='fontawesome', symbol='user')

# Current user

icon_current_user_detail = Icon(driver_name='fontawesome', symbol='user')

# Group

icon_group_create = Icon(
    driver_name='fontawesome-dual', primary_symbol='users',
    secondary_symbol='plus'
)
icon_group_detail = Icon(driver_name='fontawesome', symbol='users')
icon_group_edit = Icon(driver_name='fontawesome', symbol='pencil-alt')
icon_group_list = Icon(driver_name='fontawesome', symbol='users')
icon_group_multiple_delete = Icon(driver_name='fontawesome', symbol='times')
icon_group_setup = Icon(driver_name='fontawesome', symbol='users')
icon_group_single_delete = Icon(driver_name='fontawesome', symbol='times')
icon_group_user_list = Icon(driver_name='fontawesome', symbol='user')

# User

icon_user_create = Icon(
    driver_name='fontawesome-dual', primary_symbol='user',
    secondary_symbol='plus'
)
icon_user_edit = Icon(driver_name='fontawesome', symbol='pencil-alt')
icon_user_group_list = icon_group
icon_user_list = Icon(driver_name='fontawesome', symbol='user')
icon_user_multiple_delete = Icon(driver_name='fontawesome', symbol='times')
icon_user_multiple_set_password = Icon(
    driver_name='fontawesome', symbol='key'
)
icon_user_set_options = Icon(driver_name='fontawesome', symbol='cog')
icon_user_single_delete = Icon(driver_name='fontawesome', symbol='times')



from django import forms

from .field_mixins import FilteredModelFieldMixin


class FilteredModelChoiceField(
    FilteredModelFieldMixin, forms.ModelChoiceField
):
    """Single selection filtered model choice field"""


class FilteredModelMultipleChoiceField(
    FilteredModelFieldMixin, forms.ModelMultipleChoiceField
):
    """Multiple selection filtered model choice field"""



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('documents', '0077_favoritedocumentproxy')
    ]

    operations = [
        migrations.AlterField(
            model_name='documenttype',
            name='label',
            field=models.CharField(
                help_text='A short text identifying the document type.',
                max_length=196, unique=True, verbose_name='Label'
            )
        )
    ]



from ..models import Document, TrashedDocument

from .base import GenericDocumentTestCase


class TrashedDocumentTestCase(GenericDocumentTestCase):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_document_stub()

    def test_restoring_documents(self):
        self.assertEqual(Document.valid.count(), 1)

        # Trash the document
        self._test_document.delete()
        self.assertEqual(TrashedDocument.objects.count(), 1)
        self.assertEqual(Document.valid.count(), 0)

        # Restore the document
        TrashedDocument.objects.get(
            pk=self._test_document.pk
        ).restore(user=self._test_case_user)
        self.assertEqual(TrashedDocument.objects.count(), 0)
        self.assertEqual(Document.valid.count(), 1)

    def test_trashing_documents(self):
        self.assertEqual(Document.valid.count(), 1)

        # Trash the document
        self._test_document.delete()
        self.assertEqual(TrashedDocument.objects.count(), 1)
        self.assertEqual(Document.valid.count(), 0)

        # Delete the document
        self._test_document.delete()
        self.assertEqual(TrashedDocument.objects.count(), 0)
        self.assertEqual(Document.valid.count(), 0)


class TrashedDocumentPageTestCase(GenericDocumentTestCase):
    def test_trashed_document_page_count(self):
        page_count = self._test_document.version_active.pages.count()
        self._test_document.delete()
        test_trashed_document = TrashedDocument.objects.get(
            pk=self._test_document.pk
        )
        self.assertTrue(
            test_trashed_document.version_active.pages.count(), page_count
        )


class TrashedDocumentAPITestCase(GenericDocumentTestCase):
    def test_trashed_document_api_image_url(self):
        self._test_document.delete()
        test_trashed_document = TrashedDocument.objects.get(
            pk=self._test_document.pk
        )

        self._clear_events()

        self.assertTrue(
            test_trashed_document.get_api_image_url()
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from mayan.apps.documents.permissions import permission_document_create
from mayan.apps.documents.tests.base import GenericDocumentViewTestCase

from .mixins.base_mixins import SourceDocumentUploadViewTestMixin
from .mixins.staging_folder_source_mixins import (
    StagingFolderTestMixin, StagingFolderViewTestMixin
)


class StagingFolderActionViewTestCase(
    StagingFolderTestMixin, StagingFolderViewTestMixin,
    GenericDocumentViewTestCase
):
    auto_upload_test_document = False

    def test_staging_folder_file_delete_get_view_no_permission(self):
        self._copy_test_staging_folder_document()

        staging_folder_file_count = len(
            list(
                self._test_source.get_backend_instance().get_files()
            )
        )

        self._clear_events()

        response = self._request_staging_folder_action_file_delete_view_via_get()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            len(
                list(
                    self._test_source.get_backend_instance().get_files()
                )
            ), staging_folder_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_staging_folder_file_delete_get_view_with_access(self):
        self.grant_access(
            obj=self._test_source, permission=permission_document_create
        )
        self._copy_test_staging_folder_document()

        staging_folder_file_count = len(
            list(
                self._test_source.get_backend_instance().get_files()
            )
        )

        self._clear_events()

        response = self._request_staging_folder_action_file_delete_view_via_get()
        self.assertEqual(response.status_code, 200)

        self.assertEqual(
            len(
                list(
                    self._test_source.get_backend_instance().get_files()
                )
            ), staging_folder_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_staging_folder_file_delete_post_view_no_permission(self):
        self._copy_test_staging_folder_document()

        staging_folder_file_count = len(
            list(
                self._test_source.get_backend_instance().get_files()
            )
        )

        self._clear_events()

        response = self._request_staging_folder_action_file_delete_view_via_post()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(
            len(
                list(
                    self._test_source.get_backend_instance().get_files()
                )
            ), staging_folder_file_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_staging_folder_file_delete_post_view_with_access(self):
        self.grant_access(
            obj=self._test_source, permission=permission_document_create
        )
        self._copy_test_staging_folder_document()

        staging_folder_file_count = len(
            list(
                self._test_source.get_backend_instance().get_files()
            )
        )

        self._clear_events()

        response = self._request_staging_folder_action_file_delete_view_via_post()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(
            len(
                list(
                    self._test_source.get_backend_instance().get_files()
                )
            ), staging_folder_file_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)


class StagingFolderViewTestCase(
    StagingFolderTestMixin, SourceDocumentUploadViewTestMixin,
    GenericDocumentViewTestCase
):
    auto_create_test_source = False
    auto_upload_test_document = False

    def test_staging_folder_document_upload_view_with_full_access(self):
        self._create_test_staging_folder()

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_create
        )
        self.grant_access(
            obj=self._test_source, permission=permission_document_create
        )

        self._clear_events()

        response = self._request_source_document_upload_view_via_get()
        self.assertEqual(response.status_code, 200)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_staging_folder_document_upload_not_include_subdirectories_view_with_full_access(self):
        self._create_test_staging_folder(
            add_subdirectory=True, extra_data={
                'include_subdirectories': False
            }
        )

        self._copy_test_staging_folder_document(filename='test_document_1')
        self._copy_test_staging_folder_document(
            filename='test_document_2', to_subfolder=True
        )

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_create
        )
        self.grant_access(
            obj=self._test_source, permission=permission_document_create
        )

        self._clear_events()

        response = self._request_source_document_upload_view_via_get()
        self.assertContains(
            response=response, status_code=200, text='test_document_1'
        )
        self.assertNotContains(
            response=response, status_code=200, text='test_document_2'
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_staging_folder_document_upload_include_subdirectories_view_with_full_access(self):
        self._create_test_staging_folder(
            add_subdirectory=True, extra_data={
                'include_subdirectories': True
            }
        )

        self._copy_test_staging_folder_document(filename='test_document_1')
        self._copy_test_staging_folder_document(
            filename='test_document_2', to_subfolder=True
        )

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_create
        )
        self.grant_access(
            obj=self._test_source, permission=permission_document_create
        )

        self._clear_events()

        response = self._request_source_document_upload_view_via_get()
        self.assertContains(
            response=response, status_code=200, text='test_document_1'
        )
        self.assertContains(
            response=response, status_code=200, text='test_document_2'
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_staging_folder_document_include_regular_expression_view_with_full_access(self):
        self._create_test_staging_folder(
            extra_data={'include_regex': 'test_document_1'}
        )

        self._copy_test_staging_folder_document(filename='test_document_1')
        self._copy_test_staging_folder_document(filename='test_document_2')

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_create
        )
        self.grant_access(
            obj=self._test_source, permission=permission_document_create
        )

        self._clear_events()

        response = self._request_source_document_upload_view_via_get()
        self.assertContains(
            response=response, status_code=200, text='test_document_1'
        )
        self.assertNotContains(
            response=response, status_code=200, text='test_document_2'
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_staging_folder_document_exclude_regular_expression_view_with_full_access(self):
        self._create_test_staging_folder(
            extra_data={'exclude_regex': 'test_document_1'}
        )

        self._copy_test_staging_folder_document(filename='test_document_1')
        self._copy_test_staging_folder_document(filename='test_document_2')

        self.grant_access(
            obj=self._test_document_type,
            permission=permission_document_create
        )
        self.grant_access(
            obj=self._test_source, permission=permission_document_create
        )

        self._clear_events()

        response = self._request_source_document_upload_view_via_get()
        self.assertNotContains(
            response=response, status_code=200, text='test_document_1'
        )
        self.assertContains(
            response=response, status_code=200, text='test_document_2'
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)



from django.utils.translation import ugettext_lazy as _

from mayan.apps.documents.search import (
    search_model_document, search_model_document_file,
    search_model_document_file_page, search_model_document_version,
    search_model_document_version_page
)

# Document

search_model_document.add_model_field(
    field='comments__text', label=_('Comments')
)

# Document file

search_model_document_file.add_model_field(
    field='document__comments__text',
    label=_('Document comments')
)

# Document file page

search_model_document_file_page.add_model_field(
    field='document_file__document__comments__text',
    label=_('Document comments')
)

# Document version

search_model_document_version.add_model_field(
    field='document__comments__text',
    label=_('Document comments')
)

# Document version page

search_model_document_version_page.add_model_field(
    field='document_version__document__comments__text',
    label=_('Document comments')
)



from mayan.apps.dependencies.classes import PythonDependency

PythonDependency(
    module=__name__, name='gevent', version_string='==21.12.0'
)
PythonDependency(
    module=__name__, name='gunicorn', version_string='==20.1.0'
)
PythonDependency(
    module=__name__, name='sentry-sdk', version_string='==1.12.1'
)
PythonDependency(
    module=__name__, name='whitenoise', version_string='==6.0.0'
)



from django.utils.encoding import force_bytes

from mayan.apps.testing.tests.base import GenericViewTestCase

from ..events import event_key_created, event_key_downloaded
from ..models import Key
from ..permissions import (
    permission_key_delete, permission_key_download, permission_key_upload
)

from .literals import TEST_KEY_PRIVATE_DATA, TEST_KEY_PRIVATE_FINGERPRINT
from .mixins import KeyTestMixin, KeyViewTestMixin


class KeyViewTestCase(KeyTestMixin, KeyViewTestMixin, GenericViewTestCase):
    def test_key_delete_view_no_permission(self):
        self._create_test_key_private()

        test_key_count = Key.objects.count()

        self._clear_events()

        response = self._request_test_key_delete_view()
        self.assertEqual(response.status_code, 404)

        self.assertEqual(Key.objects.count(), test_key_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_key_delete_view_with_access(self):
        self._create_test_key_private()

        test_key_count = Key.objects.count()

        self.grant_access(
            obj=self._test_key_private, permission=permission_key_delete
        )

        self._clear_events()

        response = self._request_test_key_delete_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Key.objects.count(), test_key_count - 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_key_download_view_no_permission(self):
        self._create_test_key_private()

        self._clear_events()

        response = self._request_test_key_download_view()
        self.assertEqual(response.status_code, 404)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_key_download_view_with_access(self):
        self.expected_content_types = ('application/pgp-keys',)

        self._create_test_key_private()

        self.grant_access(
            obj=self._test_key_private, permission=permission_key_download
        )

        self._clear_events()

        response = self._request_test_key_download_view()
        self.assert_download_response(
            content=force_bytes(s=TEST_KEY_PRIVATE_DATA),
            filename=self._test_key_private.key_id, response=response
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_key_private)
        self.assertEqual(events[0].verb, event_key_downloaded.id)

    def test_key_upload_view_no_permission(self):
        self._clear_events()

        response = self._request_test_key_upload_view()
        self.assertEqual(response.status_code, 403)

        self.assertEqual(Key.objects.count(), 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_key_upload_view_with_permission(self):
        self.grant_permission(permission=permission_key_upload)

        self._clear_events()

        response = self._request_test_key_upload_view()
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Key.objects.count(), 1)
        self.assertEqual(
            Key.objects.first().fingerprint, TEST_KEY_PRIVATE_FINGERPRINT
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, None)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_key_private)
        self.assertEqual(events[0].verb, event_key_created.id)



class APIError(Exception):
    """
    Base exception for the API app.
    """


class APIResourcePatternError(APIError):
    """
    Raised when an app tries to override an existing URL regular expression
    pattern.
    """



from django.contrib.auth import get_user_model
from django.utils.translation import ugettext_lazy as _

from mayan.apps.sources.classes import DocumentCreateWizardStep
from mayan.apps.sources.wizard_steps import DocumentCreateWizardStepDocumentType

from .forms import DocumentMetadataFormSet
from .utils import (
    decode_metadata_from_query_string, save_metadata_list
)


class DocumentCreateWizardStepMetadata(DocumentCreateWizardStep):
    form_class = DocumentMetadataFormSet
    label = _('Enter document metadata')
    name = 'metadata_entry'
    number = 1

    @classmethod
    def condition(cls, wizard):
        """
        Skip step if document type has no associated metadata
        """
        cleaned_data = wizard.get_cleaned_data_for_step(
            step=DocumentCreateWizardStepDocumentType.name
        ) or {}

        document_type = cleaned_data.get('document_type')

        if document_type:
            return document_type.metadata.exists()

    @classmethod
    def get_form_initial(cls, wizard):
        initial = []

        step_data = wizard.get_cleaned_data_for_step(
            step=DocumentCreateWizardStepDocumentType.name
        )
        if step_data:
            document_type = step_data['document_type']
            for document_type_metadata_type in document_type.metadata.all():
                initial.append(
                    {
                        'document_type': document_type,
                        'metadata_type': document_type_metadata_type.metadata_type
                    }
                )

        return initial

    @classmethod
    def done(cls, wizard):
        result = {}
        cleaned_data = wizard.get_cleaned_data_for_step(step=cls.name)
        if cleaned_data:
            for index, metadata in enumerate(iterable=wizard.get_cleaned_data_for_step(step=cls.name)):
                if metadata.get('update'):
                    result[
                        'metadata{}_metadata_type_id'.format(index)
                    ] = metadata['metadata_type_id']
                    result[
                        'metadata{}_value'.format(index)
                    ] = metadata['value']

        return result

    @classmethod
    def step_post_upload_process(
        cls, document, source_id, user_id, extra_data=None, query_string=None
    ):
        User = get_user_model()

        metadata_dict_list = decode_metadata_from_query_string(
            query_string=query_string
        )

        try:
            user = User.objects.get(pk=user_id)
        except User.DoesNotExist:
            user = None

        if metadata_dict_list:
            save_metadata_list(
                create=True, document=document,
                metadata_list=metadata_dict_list, user=user
            )


DocumentCreateWizardStep.register(step=DocumentCreateWizardStepMetadata)



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('mailer', '0004_auto_20170714_2133')
    ]

    operations = [
        migrations.AlterField(
            model_name='usermailer',
            name='label',
            field=models.CharField(
                max_length=128, unique=True, verbose_name='Label'
            )
        )
    ]



from rest_framework import status

from mayan.apps.documents.tests.mixins.document_mixins import DocumentTestMixin
from mayan.apps.rest_api.tests.base import BaseAPITestCase

from ..events import (
    event_document_metadata_added, event_document_metadata_edited,
    event_document_metadata_removed
)
from ..permissions import (
    permission_document_metadata_add, permission_document_metadata_edit,
    permission_document_metadata_remove, permission_document_metadata_view
)

from .literals import TEST_METADATA_TYPE_DEFAULT_VALUE
from .mixins import (
    DocumentMetadataAPIViewTestMixin, DocumentMetadataMixin,
    MetadataTypeTestMixin
)


class DocumentMetadataAPIViewTestCase(
    DocumentMetadataAPIViewTestMixin, DocumentMetadataMixin,
    DocumentTestMixin, MetadataTypeTestMixin, BaseAPITestCase
):
    auto_upload_test_document = False

    def setUp(self):
        super().setUp()
        self._create_test_document_stub()
        self._create_test_metadata_type()
        self._test_metadata_type.default = TEST_METADATA_TYPE_DEFAULT_VALUE
        self._test_metadata_type.save()
        self._test_document_type.metadata.create(
            metadata_type=self._test_metadata_type, required=False
        )

    def test_document_metadata_create_api_view_no_permission(self):
        document_metadata_count = self._test_document.metadata.count()

        self._clear_events()

        response = self._request_document_metadata_create_api_view()

        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            self._test_document.metadata.count(), document_metadata_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_create_api_view_with_document_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_add
        )

        document_metadata_count = self._test_document.metadata.count()

        self._clear_events()

        response = self._request_document_metadata_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(
            self._test_document.metadata.count(), document_metadata_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_create_api_view_with_metadata_type_access(self):
        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_add
        )

        document_metadata_count = self._test_document.metadata.count()

        self._clear_events()

        response = self._request_document_metadata_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            self._test_document.metadata.count(), document_metadata_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_create_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_add
        )
        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_add
        )

        document_metadata_count = self._test_document.metadata.count()

        self._clear_events()

        response = self._request_document_metadata_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(
            self._test_document.metadata.count(), document_metadata_count + 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_metadata_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_metadata_added.id)

    def test_trashed_document_metadata_create_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_add
        )
        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_add
        )

        document_metadata_count = self._test_document.metadata.count()

        self._test_document.delete()

        self._clear_events()

        response = self._request_document_metadata_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            self._test_document.metadata.count(), document_metadata_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_create_default_value_api_view_with_full_access(self):
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_add
        )
        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_add
        )

        self._clear_events()

        response = self._request_document_metadata_create_api_view(
            extra_data={'value': ''}
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        document_metadata = self._test_document.metadata.first()
        self.assertEqual(response.data['id'], document_metadata.pk)
        self.assertEqual(
            document_metadata.metadata_type, self._test_metadata_type
        )
        self.assertEqual(
            document_metadata.value, TEST_METADATA_TYPE_DEFAULT_VALUE
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_metadata_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_metadata_added.id)

    def test_document_metadata_create_duplicate_api_view(self):
        self._create_test_document_metadata()
        self.grant_permission(permission=permission_document_metadata_add)

        self._clear_events()

        response = self._request_document_metadata_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(list(response.data.keys())[0], 'non_field_errors')

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_create_invalid_lookup_value_api_view(self):
        self._test_metadata_type.lookup = 'invalid,lookup,values,on,purpose'
        self._test_metadata_type.save()
        self.grant_permission(permission=permission_document_metadata_add)

        self._clear_events()

        response = self._request_document_metadata_create_api_view()
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(list(response.data.keys())[0], 'non_field_errors')

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_delete_api_view_no_permission(self):
        self._create_test_document_metadata()

        document_metadata_count = self._test_document.metadata.count()

        self._clear_events()

        response = self._request_document_metadata_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            self._test_document.metadata.count(), document_metadata_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_delete_api_view_with_document_access(self):
        self._create_test_document_metadata()
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_remove
        )

        document_metadata_count = self._test_document.metadata.count()

        self._clear_events()

        response = self._request_document_metadata_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            self._test_document.metadata.count(), document_metadata_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_delete_api_view_with_metadata_type_access(self):
        self._create_test_document_metadata()
        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_remove
        )

        document_metadata_count = self._test_document.metadata.count()

        self._clear_events()

        response = self._request_document_metadata_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            self._test_document.metadata.count(), document_metadata_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_delete_api_view_with_full_access(self):
        self._create_test_document_metadata()
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_remove
        )
        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_remove
        )

        document_metadata_count = self._test_document.metadata.count()

        self._clear_events()

        response = self._request_document_metadata_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertEqual(
            self._test_document.metadata.count(), document_metadata_count - 1
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_metadata_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_metadata_removed.id)

    def test_trashed_document_metadata_delete_api_view_with_full_access(self):
        self._create_test_document_metadata()
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_remove
        )
        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_remove
        )

        document_metadata_count = self._test_document.metadata.count()

        self._test_document.delete()

        self._clear_events()

        response = self._request_document_metadata_delete_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(
            self._test_document.metadata.count(), document_metadata_count
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_list_api_view_no_permission(self):
        self._create_test_document_metadata()

        self._clear_events()

        response = self._request_document_metadata_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_list_api_view_with_document_access(self):
        self._create_test_document_metadata()

        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_view
        )

        self._clear_events()

        response = self._request_document_metadata_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['count'], 0)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_list_api_view_with_metadata_type_access(self):
        self._create_test_document_metadata()

        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_view
        )

        self._clear_events()

        response = self._request_document_metadata_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_list_api_view_with_full_access(self):
        self._create_test_document_metadata()
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_view
        )
        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_view
        )

        self._clear_events()

        response = self._request_document_metadata_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(
            response.data['results'][0]['document']['id'],
            self._test_document.pk
        )
        self.assertEqual(
            response.data['results'][0]['metadata_type']['id'],
            self._test_metadata_type.pk
        )
        self.assertEqual(
            response.data['results'][0]['value'],
            self._test_document_metadata.value
        )
        self.assertEqual(
            response.data['results'][0]['id'],
            self._test_document_metadata.pk
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_trashed_document_metadata_list_api_view_with_full_access(self):
        self._create_test_document_metadata()
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_view
        )
        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_view
        )

        self._test_document.delete()

        self._clear_events()

        response = self._request_document_metadata_list_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_patch_api_view_no_permission(self):
        self._create_test_document_metadata()

        document_metadata_value = self._test_document_metadata.value

        self._clear_events()

        response = self._request_document_metadata_edit_api_view_via_patch()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_document_metadata.refresh_from_db()
        self.assertEqual(
            self._test_document_metadata.value, document_metadata_value
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_patch_api_view_document_access(self):
        self._create_test_document_metadata()
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_edit
        )

        document_metadata_value = self._test_document_metadata.value

        self._clear_events()

        response = self._request_document_metadata_edit_api_view_via_patch()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_document_metadata.refresh_from_db()
        self.assertEqual(
            self._test_document_metadata.value, document_metadata_value
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_patch_api_view_metadata_type_access(self):
        self._create_test_document_metadata()
        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_edit
        )

        document_metadata_value = self._test_document_metadata.value

        self._clear_events()

        response = self._request_document_metadata_edit_api_view_via_patch()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_document_metadata.refresh_from_db()
        self.assertEqual(
            self._test_document_metadata.value, document_metadata_value
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_patch_api_view_with_full_access(self):
        self._create_test_document_metadata()
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_edit
        )
        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_edit
        )

        document_metadata_value = self._test_document_metadata.value

        self._clear_events()

        response = self._request_document_metadata_edit_api_view_via_patch()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self._test_document_metadata.refresh_from_db()
        self.assertNotEqual(
            self._test_document_metadata.value, document_metadata_value
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_metadata_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_metadata_edited.id)

    def test_trashed_document_metadata_patch_api_view_with_full_access(self):
        self._create_test_document_metadata()
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_edit
        )
        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_edit
        )

        document_metadata_value = self._test_document_metadata.value

        self._test_document.delete()

        self._clear_events()

        response = self._request_document_metadata_edit_api_view_via_patch()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_document_metadata.refresh_from_db()
        self.assertEqual(
            self._test_document_metadata.value, document_metadata_value
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_patch_default_value_api_view_with_full_access(self):
        self._create_test_document_metadata()
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_edit
        )
        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_edit
        )

        self._clear_events()

        response = self._request_document_metadata_edit_api_view_via_patch(
            extra_data={'value': ''}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self._test_document_metadata.refresh_from_db()
        self.assertEqual(
            response.data['value'], TEST_METADATA_TYPE_DEFAULT_VALUE
        )
        self.assertEqual(
            self._test_document_metadata.value,
            TEST_METADATA_TYPE_DEFAULT_VALUE
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_metadata_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_metadata_edited.id)

    def test_document_metadata_put_api_view_no_permission(self):
        self._create_test_document_metadata()

        document_metadata_value = self._test_document_metadata.value

        self._clear_events()

        response = self._request_document_metadata_edit_api_view_via_put()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_document_metadata.refresh_from_db()
        self.assertEqual(
            self._test_document_metadata.value, document_metadata_value
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_put_api_view_document_access(self):
        self._create_test_document_metadata()
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_edit
        )

        document_metadata_value = self._test_document_metadata.value

        self._clear_events()

        response = self._request_document_metadata_edit_api_view_via_put()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_document_metadata.refresh_from_db()
        self.assertEqual(
            self._test_document_metadata.value, document_metadata_value
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_put_api_view_metadata_type_access(self):
        self._create_test_document_metadata()
        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_edit
        )

        document_metadata_value = self._test_document_metadata.value

        self._clear_events()

        response = self._request_document_metadata_edit_api_view_via_put()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_document_metadata.refresh_from_db()
        self.assertEqual(
            self._test_document_metadata.value, document_metadata_value
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_put_api_view_with_full_access(self):
        self._create_test_document_metadata()
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_edit
        )
        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_edit
        )

        document_metadata_value = self._test_document_metadata.value

        self._clear_events()

        response = self._request_document_metadata_edit_api_view_via_put()
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self._test_document_metadata.refresh_from_db()
        self.assertNotEqual(
            self._test_document_metadata.value, document_metadata_value
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_metadata_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_metadata_edited.id)

    def test_trashed_document_metadata_put_api_view_with_full_access(self):
        self._create_test_document_metadata()
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_edit
        )
        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_edit
        )

        document_metadata_value = self._test_document_metadata.value

        self._test_document.delete()

        self._clear_events()

        response = self._request_document_metadata_edit_api_view_via_put()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self._test_document_metadata.refresh_from_db()
        self.assertEqual(
            self._test_document_metadata.value, document_metadata_value
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_document_metadata_put_default_value_api_view_with_full_access(self):
        self._create_test_document_metadata()
        self.grant_access(
            obj=self._test_document,
            permission=permission_document_metadata_edit
        )
        self.grant_access(
            obj=self._test_metadata_type,
            permission=permission_document_metadata_edit
        )

        self._clear_events()

        response = self._request_document_metadata_edit_api_view_via_put(
            extra_data={'value': ''}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

        self._test_document_metadata.refresh_from_db()
        self.assertEqual(
            response.data['value'], TEST_METADATA_TYPE_DEFAULT_VALUE
        )
        self.assertEqual(
            self._test_document_metadata.value,
            TEST_METADATA_TYPE_DEFAULT_VALUE
        )

        events = self._get_test_events()
        self.assertEqual(events.count(), 1)

        self.assertEqual(events[0].action_object, self._test_metadata_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, self._test_document)
        self.assertEqual(events[0].verb, event_document_metadata_edited.id)



import logging

from django.core.files.base import ContentFile
from django.db import models
from django.template.defaultfilters import filesizeformat
from django.utils.translation import ugettext_lazy as _, ugettext

from mayan.apps.events.classes import EventManagerMethodAfter
from mayan.apps.events.decorators import method_event

from .events import event_download_file_downloaded

logger = logging.getLogger(name=__name__)


class DatabaseFileModelMixin(models.Model):
    filename = models.CharField(
        db_index=True, max_length=255, verbose_name=_('Filename')
    )
    datetime = models.DateTimeField(
        auto_now_add=True, verbose_name=_('Date time')
    )

    class Meta:
        abstract = True

    def delete(self, *args, **kwargs):
        name = self.file.name
        self.file.close()

        if name:
            self.file.storage.delete(name=name)
        return super().delete(*args, **kwargs)

    def open(self, **kwargs):
        default_kwargs = {
            'mode': getattr(self.file.file, 'mode', 'rb'),
            'name': self.file.name
        }

        # Close the self.file object as Django generates a new descriptor
        # when the file field is accessed.
        # From django/db/models/fields/files.py.
        """
        The descriptor for the file attribute on the model instance. Return a
        FieldFile when accessed so you can write code like::

            >>> from myapp.models import MyModel
            >>> instance = MyModel.objects.get(pk=1)
            >>> instance.file.size

        Assign a file object on assignment so you can do::

            >>> with open('/path/to/hello.world', 'r') as f:
            ...     instance.file = File(f)
        """
        self.file.close()
        self.file.file.close()

        default_kwargs.update(**kwargs)

        # Ensure the caller cannot specify an alternate filename.
        name = kwargs.pop('name', None)

        if name:
            logger.warning(
                'Caller tried to specify an alternate filename: %s', name
            )

        return self.file.storage.open(**default_kwargs)

    def save(self, *args, **kwargs):
        if not self.file:
            self.file = ContentFile(
                content=b'', name=self.filename or ugettext('Unnamed file')
            )

        self.filename = self.filename or str(self.file)
        super().save(*args, **kwargs)


class DownloadFileBusinessLogicMixin:
    @method_event(
        event_manager_class=EventManagerMethodAfter,
        event=event_download_file_downloaded,
        target='self'
    )
    def get_download_file_object(self):
        return self.open(mode='rb')

    def get_size_display(self):
        return filesizeformat(bytes_=self.file.size)

    get_size_display.short_description = _('Size')

    def get_user_display(self):
        if self.user.get_full_name():
            return self.user.get_full_name()
        else:
            return self.user.username
    get_user_display.short_description = _('User')



from rest_framework import status

from mayan.apps.documents.events import (
    event_document_created, event_document_file_created,
    event_document_file_edited, event_document_version_created,
    event_document_version_page_created
)
from mayan.apps.documents.models.document_models import Document
from mayan.apps.documents.permissions import permission_document_create
from mayan.apps.documents.tests.literals import TEST_FILE_COMPRESSED_PATH
from mayan.apps.documents.tests.mixins.document_mixins import DocumentTestMixin
from mayan.apps.rest_api.tests.base import BaseAPITestCase

from ..source_backends.literals import (
    SOURCE_UNCOMPRESS_CHOICE_ASK, SOURCE_UNCOMPRESS_CHOICE_NEVER
)

from .mixins.web_form_source_mixins import (
    WebFormSourceBackendAPITestMixin, WebFormSourceBackendTestMixin
)


class WebFormSourceBackendAPITestCase(
    WebFormSourceBackendTestMixin, WebFormSourceBackendAPITestMixin,
    DocumentTestMixin, BaseAPITestCase
):
    auto_create_test_source = False
    auto_upload_test_document = False

    def test_web_form_file_upload_api_view_with_no_permission(self):
        self._create_test_web_form_source(
            extra_data={'uncompress': SOURCE_UNCOMPRESS_CHOICE_NEVER}
        )

        document_count = Document.objects.count()

        self._clear_events()

        response = self._request_test_web_form_file_upload_action_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(Document.objects.count(), document_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_web_form_file_upload_api_view_with_document_type_access(self):
        self._create_test_web_form_source(
            extra_data={'uncompress': SOURCE_UNCOMPRESS_CHOICE_NEVER}
        )

        self.grant_access(
            obj=self._test_document_type, permission=permission_document_create
        )

        document_count = Document.objects.count()

        self._clear_events()

        response = self._request_test_web_form_file_upload_action_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(Document.objects.count(), document_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_web_form_file_upload_api_view_with_source_access(self):
        self._create_test_web_form_source(
            extra_data={'uncompress': SOURCE_UNCOMPRESS_CHOICE_NEVER}
        )

        self.grant_access(
            obj=self._test_source, permission=permission_document_create
        )

        document_count = Document.objects.count()

        self._clear_events()

        response = self._request_test_web_form_file_upload_action_api_view()
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

        self.assertEqual(Document.objects.count(), document_count)

        events = self._get_test_events()
        self.assertEqual(events.count(), 0)

    def test_web_form_file_upload_api_view_with_full_access(self):
        self._create_test_web_form_source(
            extra_data={'uncompress': SOURCE_UNCOMPRESS_CHOICE_NEVER}
        )

        self.grant_access(
            obj=self._test_document_type, permission=permission_document_create
        )
        self.grant_access(
            obj=self._test_source, permission=permission_document_create
        )

        document_count = Document.objects.count()

        self._clear_events()

        response = self._request_test_web_form_file_upload_action_api_view()
        self.assertEqual(response.status_code, status.HTTP_202_ACCEPTED)

        self.assertEqual(Document.objects.count(), document_count + 1)

        events = self._get_test_events()
        self.assertEqual(events.count(), 5)

        test_document = Document.objects.first()
        test_document_file = test_document.file_latest
        test_document_version = test_document.version_active
        test_document_version_page = test_document_version.pages.first()

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, test_document)
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_document)
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_document_file)
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_document)
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_document_file)
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, test_document)
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, test_document_version)
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(events[4].action_object, test_document_version)
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(events[4].target, test_document_version_page)
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

    def test_web_form_compressed_file_upload_api_view_with_full_access(self):
        self._create_test_web_form_source(
            extra_data={'uncompress': SOURCE_UNCOMPRESS_CHOICE_ASK}
        )

        self.grant_access(
            obj=self._test_document_type, permission=permission_document_create
        )
        self.grant_access(
            obj=self._test_source, permission=permission_document_create
        )

        document_count = Document.objects.count()

        self._clear_events()

        response = self._request_test_web_form_file_upload_action_api_view(test_file_path=TEST_FILE_COMPRESSED_PATH)
        self.assertEqual(response.status_code, status.HTTP_202_ACCEPTED)

        self.assertEqual(Document.objects.count(), document_count + 2)

        events = self._get_test_events()
        self.assertEqual(events.count(), 11)

        test_documents = (Document.objects.first(), Document.objects.last())
        test_document_files = (
            test_documents[0].file_latest, test_documents[1].file_latest
        )
        test_document_versions = (
            test_documents[0].version_active, test_documents[1].version_active
        )
        test_document_version_pages = (
            test_document_versions[0].pages.first(),
            test_document_versions[1].pages.first(),
            test_document_versions[1].pages.last()
        )

        self.assertEqual(events[0].action_object, self._test_document_type)
        self.assertEqual(events[0].actor, self._test_case_user)
        self.assertEqual(events[0].target, test_documents[0])
        self.assertEqual(events[0].verb, event_document_created.id)

        self.assertEqual(events[1].action_object, test_documents[0])
        self.assertEqual(events[1].actor, self._test_case_user)
        self.assertEqual(events[1].target, test_document_files[0])
        self.assertEqual(events[1].verb, event_document_file_created.id)

        self.assertEqual(events[2].action_object, test_documents[0])
        self.assertEqual(events[2].actor, self._test_case_user)
        self.assertEqual(events[2].target, test_document_files[0])
        self.assertEqual(events[2].verb, event_document_file_edited.id)

        self.assertEqual(events[3].action_object, test_documents[0])
        self.assertEqual(events[3].actor, self._test_case_user)
        self.assertEqual(events[3].target, test_document_versions[0])
        self.assertEqual(events[3].verb, event_document_version_created.id)

        self.assertEqual(events[4].action_object, test_document_versions[0])
        self.assertEqual(events[4].actor, self._test_case_user)
        self.assertEqual(events[4].target, test_document_version_pages[0])
        self.assertEqual(
            events[4].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[5].action_object, self._test_document_type)
        self.assertEqual(events[5].actor, self._test_case_user)
        self.assertEqual(events[5].target, test_documents[1])
        self.assertEqual(events[5].verb, event_document_created.id)

        self.assertEqual(events[6].action_object, test_documents[1])
        self.assertEqual(events[6].actor, self._test_case_user)
        self.assertEqual(events[6].target, test_document_files[1])
        self.assertEqual(events[6].verb, event_document_file_created.id)

        self.assertEqual(events[7].action_object, test_documents[1])
        self.assertEqual(events[7].actor, self._test_case_user)
        self.assertEqual(events[7].target, test_document_files[1])
        self.assertEqual(events[7].verb, event_document_file_edited.id)

        self.assertEqual(events[8].action_object, test_documents[1])
        self.assertEqual(events[8].actor, self._test_case_user)
        self.assertEqual(events[8].target, test_document_versions[1])
        self.assertEqual(events[8].verb, event_document_version_created.id)

        self.assertEqual(events[9].action_object, test_document_versions[1])
        self.assertEqual(events[9].actor, self._test_case_user)
        self.assertEqual(events[9].target, test_document_version_pages[1])
        self.assertEqual(
            events[9].verb, event_document_version_page_created.id
        )

        self.assertEqual(events[10].action_object, test_document_versions[1])
        self.assertEqual(events[10].actor, self._test_case_user)
        self.assertEqual(events[10].target, test_document_version_pages[2])
        self.assertEqual(
            events[10].verb, event_document_version_page_created.id
        )



from django import forms
from django.utils.translation import ugettext_lazy as _

from .literals import MATCH_ALL_FIELD_CHOICES, MATCH_ALL_FIELD_NAME


class AdvancedSearchForm(forms.Form):
    def __init__(self, *args, **kwargs):
        kwargs['data'] = kwargs['data'].copy()
        self.search_model = kwargs.pop('search_model')
        super().__init__(*args, **kwargs)

        self.fields[MATCH_ALL_FIELD_NAME] = forms.ChoiceField(
            choices=MATCH_ALL_FIELD_CHOICES, widget=forms.RadioSelect,
            label=_('Match all'), help_text=_(
                'Return only results that match all fields.'
            ), required=False
        )

        for search_field in self.search_model.search_fields:
            if search_field.concrete:
                self.fields[search_field.field_name] = forms.CharField(
                    help_text=search_field.get_help_text(),
                    label=search_field.label, required=False,
                    widget=forms.widgets.TextInput(
                        attrs={'type': 'search'}
                    )
                )


class SearchForm(forms.Form):
    q = forms.CharField(
        max_length=128, label=_('Search terms'), required=False,
        widget=forms.widgets.TextInput(
            attrs={'type': 'search'}
        )
    )



from ..document_file_actions import (
    DocumentFileActionAppendNewPages, DocumentFileActionNothing,
    DocumentFileActionUseNewPages
)

from .base import GenericDocumentTestCase
from .mixins.document_file_mixins import DocumentFileTestMixin


class DocumentVersionTestCase(
    DocumentFileTestMixin, GenericDocumentTestCase
):
    def test_version_new_file_new_pages(self):
        test_document_version_page_content_objects = self._test_document_version.page_content_objects

        self.assertEqual(self._test_document.versions.count(), 1)

        self._upload_test_document_file(
            action=DocumentFileActionUseNewPages.backend_id
        )

        self.assertEqual(self._test_document.versions.count(), 2)

        self.assertNotEqual(
            self._test_document_version.page_content_objects,
            test_document_version_page_content_objects
        )
        self.assertEqual(
            self._test_document_version.page_content_objects,
            list(self._test_document.file_latest.pages.all())
        )

    def test_version_new_version_keep_pages(self):
        test_document_version_page_content_objects = self._test_document_version.page_content_objects

        self.assertEqual(self._test_document.versions.count(), 1)

        self._upload_test_document_file(
            action=DocumentFileActionNothing.backend_id
        )

        self.assertEqual(self._test_document.versions.count(), 1)

        self.assertEqual(
            self._test_document_version.page_content_objects,
            test_document_version_page_content_objects
        )
        self.assertNotEqual(
            self._test_document_version.page_content_objects,
            list(self._test_document.file_latest.pages.all())
        )

    def test_version_new_file_append_pages(self):
        test_document_version_page_content_objects = self._test_document_version.page_content_objects

        self.assertEqual(self._test_document.versions.count(), 1)
        self.assertEqual(self._test_document.files.count(), 1)

        self._upload_test_document_file(
            action=DocumentFileActionAppendNewPages.backend_id
        )

        self.assertEqual(self._test_document.files.count(), 2)
        self.assertEqual(self._test_document.versions.count(), 2)

        test_document_version_expected_page_content_objects = list(
            self._test_document.files.first().pages.all()
        )
        test_document_version_expected_page_content_objects.extend(
            list(
                self._test_document.files.last().pages.all()
            )
        )

        self.assertNotEqual(
            self._test_document_version.page_content_objects,
            test_document_version_page_content_objects
        )
        self.assertEqual(
            self._test_document_version.page_content_objects,
            test_document_version_expected_page_content_objects
        )

    def test_method_get_absolute_url(self):
        self.assertTrue(
            self._test_document.version_active.get_absolute_url()
        )



from django.core.management.base import BaseCommand

from ...models import UserOTPData


class Command(BaseCommand):
    help = 'Display the OTP status for a user.'
    missing_args_message = 'You must provide a username.'

    def add_arguments(self, parser):
        parser.add_argument(
            dest='username', help='Username for which OTP status will be displayed.',
            metavar='<username>'
        )

    def handle(self, username, **options):
        try:
            otp_data = UserOTPData.objects.get(user__username=username)
        except UserOTPData.DoesNotExist:
            status_text = 'Unknown username or OTP data is not initialized correctly.'
        else:
            if otp_data.is_enabled():
                status_text = 'OTP is enabled.'
            else:
                status_text = 'OTP is disabled.'

        self.stdout.write(msg='OTP state is: {}'.format(status_text))
